"""Test patch utilities."""

from __future__ import annotations

from typing import Annotated, Literal

import numpy as np
import pandas as pd
import pydantic
import pytest
from pydantic import Field

import dascore as dc
from dascore import patch_function
from dascore.constants import PatchType
from dascore.exceptions import (
    IncompatiblePatchError,
    ParameterError,
    PatchAttributeError,
    PatchCoordinateError,
)
from dascore.utils.patch import (
    _spool_up,
    align_patch_coords,
    concatenate_patches,
    get_dim_axis_value,
    get_patch_names,
    merge_compatible_coords_attrs,
    patches_to_df,
    stack_patches,
)


@dc.patch_function(required_dims=("time", "distance"))
def time_dist_dims(patch):
    """A dummy function to test time, distance dim requirement."""
    return patch


@dc.patch_function()
def add_one(patch):
    """A simple function which modifies the patch."""
    return patch.new(data=patch.data + 1)


@dc.patch_function(required_coords=("time", "distance"))
def time_dist_coord(patch):
    """A dummy function to test time, distance coord requirement."""
    return patch


@dc.patch_function(required_attrs=("bob",))
def require_bob(patch):
    """Require bob attribute."""
    return patch


@dc.patch_function(required_attrs={"bob": "what about?"})
def require_bob_value(patch):
    """Require bob attribute with specific value."""
    return patch


class TestPatchFunction:
    """Tests for patch function decorator."""

    def test_no_call(self, random_patch):
        """Ensure the decorator can be used with or without parens."""

        @dc.patch_function
        def func1(patch):
            """First test func."""
            return patch

        @dc.patch_function()
        def func2(patch):
            """Second test func."""
            return patch

        pa1 = func1(random_patch)
        pa2 = func2(random_patch)
        assert pa1.equals(pa2)

    def test_has_required_dims(self, random_patch):
        """Test that the required dimension check works."""
        # passes if this doesn't raise
        time_dist_dims(random_patch)

    def test_doesnt_have_required_dims(self, random_patch):
        """Raises if patch doesn't have required dimensions."""
        pa = random_patch.rename_coords(time="for_a_hug")

        with pytest.raises(PatchCoordinateError):
            time_dist_dims(pa)

    def test_has_required_coords(self, random_patch, random_dft_patch):
        """Test that the required coord check works."""
        # if the coords are dimensions it should work
        time_dist_coord(random_patch)
        # or if they are only coordinates it should work
        time_dist_coord(random_dft_patch)

    def test_doesnt_have_required_coords(self, random_patch):
        """Raises if patch doesn't have required coords."""
        pa = random_patch.rename_coords(time="for_a_hug")
        with pytest.raises(PatchCoordinateError):
            time_dist_coord(pa)

    def test_required_attrs(self, random_patch):
        """Tests for requiring certain attrs."""
        with pytest.raises(PatchAttributeError):
            require_bob(random_patch)

    def test_require_attr_exists(self, random_patch):
        """Test for requiring an attr exists but not checking value."""
        pa1 = random_patch.update_attrs(bob=1)
        pa2 = pa1.update_attrs(bob=2)
        # both these should not raise
        require_bob(pa1)
        require_bob(pa2)

    def test_require_attr_and_value(self, random_patch):
        """Tests for requiring attribute and specific value."""
        pa1 = random_patch.update_attrs(bob="what about?")
        pa2 = random_patch.update_attrs(bob="lightening")
        require_bob_value(pa1)  # should not raise
        with pytest.raises(PatchAttributeError, match="lightening"):
            require_bob_value(pa2)

    def test_access_original_function(self, random_patch):
        """Ensure the original function is still accessible."""
        hist1 = ",".join(random_patch.attrs["history"])
        # this calls the original function so it should bypass history logging
        out = require_bob.func(random_patch)
        hist2 = ",".join(out.attrs["history"])
        assert hist1 == hist2

    def test_type_checking(self, random_patch):
        """Ensure the type-checking for validate_call works."""

        @patch_function(validate_call=True)
        def some_func(
            patch: PatchType,
            some_int: int,
            specific_float: Annotated[float, Field(ge=0, le=1)],
            lit_str: Literal["bob", "bill", "marely"] = "bill",
        ):
            """A test function for type checking"""
            return patch

        patch = dc.get_example_patch()
        ok = some_func(patch, some_int=23, specific_float=0.2, lit_str="bob")
        assert isinstance(ok, dc.Patch)
        assert ok is patch

        with pytest.raises(pydantic.ValidationError):
            some_func(patch, some_int=10, specific_float=20.0)


class TestHistory:
    """Tests for tracking patch processing history."""

    def test_simple_history(self, random_patch):
        """Just make sure function is logged."""
        out = add_one(random_patch)
        history = out.attrs["history"]
        assert len(history) == 1
        assert "add_one" in history[0]

    def test_original_history_unchanged(self, random_patch):
        """Ensure logging history only occurs on new Patch."""
        first_history = list(random_patch.attrs["history"])
        _ = add_one(random_patch)
        last_history = list(random_patch.attrs["history"])
        assert first_history == last_history

    def test_no_history_unchanged_patch(self, random_patch):
        """
        Ensure no new history is created by a function that returns the same patch.
        """
        out = time_dist_dims(random_patch)
        assert out.attrs.history == random_patch.attrs.history
        # No new patch should have been created.
        assert out is random_patch


class TestMergePatches:
    """Tests for merging patches together."""

    def test_deprecated(self, random_patch):
        """Ensure deprecation warning is raised."""
        from dascore.utils.patch import merge_patches

        with pytest.warns(DeprecationWarning, match="merge_patches is deprecated"):
            merge_patches(random_patch)


class TestGetDimAxisValue:
    """Tests for getting the dimension name, axis, and value."""

    def test_raises_no_overlap(self, random_patch):
        """Test that an exception is raised when key doesn't exist."""
        kwargs = {}
        msg = "exactly one dimension"
        with pytest.raises(PatchCoordinateError, match=msg):
            get_dim_axis_value(random_patch, kwargs=kwargs)

    def test_raises_extra(self, random_patch):
        """Ensure extra args/kwargs raise."""
        msg = "not found in the patch"
        with pytest.raises(PatchCoordinateError, match=msg):
            get_dim_axis_value(random_patch, kwargs={"bob": 10, "time": None})
        with pytest.raises(PatchCoordinateError, match=msg):
            get_dim_axis_value(random_patch, args=("bob", "distance"))

    def test_get_single(self, random_patch):
        """Ensure a single kwarg works."""
        value = 10
        for axis, dim in enumerate(random_patch.dims):
            # test kwargs
            out = get_dim_axis_value(random_patch, kwargs={dim: value})
            assert len(out) == 1
            assert out[0] == (dim, axis, value)
            # test args
            out = get_dim_axis_value(random_patch, args=(dim,))
            assert len(out) == 1
            assert out[0] == (dim, axis, None)

    def test_multiple(self, random_patch):
        """Ensure multiple kwargs works."""
        kwargs = {x: 10 for x in random_patch.dims}
        # Allow multiple should return a tuple.
        out = get_dim_axis_value(random_patch, kwargs=kwargs, allow_multiple=True)
        assert len(out) == len(random_patch.dims)
        # But if not it should raise
        msg = "exactly one dimension"
        with pytest.raises(PatchCoordinateError, match=msg):
            get_dim_axis_value(random_patch, kwargs=kwargs, allow_multiple=False)


class TestPatchesToDF:
    """Test for getting metadata from patch into a dataframe."""

    def test_spool_input(self, random_spool):
        """A spool should return its contents."""
        df = patches_to_df(random_spool)
        assert isinstance(df, pd.DataFrame)
        assert len(df) == len(random_spool)

    def test_dataframe_input(self, random_spool):
        """The function should be idempotent."""
        df = random_spool.get_contents()
        out = patches_to_df(df)
        out2 = patches_to_df(out)
        eq = out == out2
        is_null = pd.isnull(out) & pd.isnull(out2)
        assert np.all(eq | is_null)

    def test_history_added(self, random_spool):
        """Ensure the history column gets added."""
        df = random_spool.get_contents().drop(columns="history", errors="ignore")
        out = patches_to_df(df)
        assert "history" in out.columns


class TestAlignPatches:
    """Tests for aligning patches."""

    def test_align_self(self, random_patch):
        """A patch should align with itself."""
        out1, out2 = align_patch_coords(random_patch, random_patch)
        assert out1 == out2 == random_patch

    def test_different_dims(self, random_patch):
        """Ensure alignment works for different dimensions."""
        no_time = random_patch.select(time=0, samples=True).squeeze()
        out1, out2 = align_patch_coords(no_time, random_patch)
        assert out1.dims == out2.dims
        assert out1.ndim == out2.ndim
        eq_or_q = [x == y or 1 in {x, y} for x, y in zip(out1.shape, out2.shape)]
        assert all(eq_or_q)

    def test_subset(self, random_patch):
        """Ensure a subset of a patch slices output."""
        sub = random_patch.select(time=(1, 15), samples=True)
        out1, out2 = align_patch_coords(sub, random_patch)
        assert out1.shape == out2.shape
        assert out1.coords == out2.coords
        # The data should all be the same since this is a sub-patch.
        assert np.all(out1.data == out2.data)

    def test_align_no_overlap(self, random_patch):
        """Alignment with no overlap should return null."""
        dist = random_patch.get_coord("distance")
        new = random_patch.update_coords(distance_min=dist.max() + 10)
        out1, out2 = align_patch_coords(new, random_patch)
        assert out1.shape == out2.shape
        assert out1.coords == out2.coords
        # Patches should be degenerate.
        assert 0 in set(out1.shape)

    def test_no_common_dims_raises(self, random_patch):
        """Patches with no common dims should not be align-able."""
        new = random_patch.rename_coords(time="money", distance="gold")
        msg = "align patches with no shared dimensions."
        with pytest.raises(PatchCoordinateError, match=msg):
            align_patch_coords(new, random_patch)

    def test_new_dims_each_patch(self, random_patch):
        """Tests for when there are new dimensions on each patch."""
        small = random_patch.select(time=(0, 10), distance=(0, 12), samples=True)
        p1 = small.rename_coords(time="money")
        p2 = small.rename_coords(time="about")
        out1, out2 = align_patch_coords(p1, p2)
        assert out1.dims == out2.dims
        assert out1.ndim == out2.ndim

    def test_len_1_non_coord_patch(self, random_patch):
        """Tests for when there are non-coordinate dimensions of len 1."""
        non_patch = random_patch.first("time")
        # Coords should be expanded in out2.
        out1, out2 = align_patch_coords(non_patch, random_patch)
        desired_shape = tuple(max(x, y) for x, y in zip(out1.shape, out2.shape))
        # This will raise if the shapes aren't broadcastable
        np.broadcast_shapes(out1.shape, out2.shape, desired_shape)


class TestMergeCompatibleCoordsAttrs:
    """Tests for merging compatible attrs, coords."""

    def test_simple(self, random_patch):
        """Simple merge test."""
        coords, attrs = merge_compatible_coords_attrs(random_patch, random_patch)
        assert coords == random_patch.coords
        assert attrs == random_patch.attrs

    def test_incompatible_dims(self, random_patch):
        """Ensure incompatible dims raises."""
        new = random_patch.rename_coords(time="money")
        match = "not compatible for merging"
        with pytest.raises(IncompatiblePatchError, match=match):
            merge_compatible_coords_attrs(random_patch, new)

    def test_incompatible_coords(self, random_patch):
        """Ensure an incompatible error is raised for coords that dont match."""
        new_time = random_patch.attrs.time_max
        new = random_patch.update_attrs(time_min=new_time)
        match = "coordinates are not equal"
        with pytest.raises(IncompatiblePatchError, match=match):
            merge_compatible_coords_attrs(new, random_patch)

    def test_incompatible_attrs(self, random_patch):
        """Ensure if attrs are off an Error is raised."""
        new = random_patch.update_attrs(network="TA")
        match = "attributes are not equal"
        with pytest.raises(IncompatiblePatchError, match=match):
            merge_compatible_coords_attrs(new, random_patch)

    def test_extra_coord(self, random_patch, random_patch_with_lat_lon):
        """Extra coords on both patch should end up in the merged patch."""
        new_coord = np.ones(random_patch.coord_shapes["time"])
        pa1 = random_patch.update_coords(new_time=("time", new_coord))
        pa2 = random_patch_with_lat_lon.update_coords(new_time=("time", new_coord))
        expected = set(pa1.coords.coord_map) | set(pa2.coords.coord_map)
        coords, attrs = merge_compatible_coords_attrs(pa1, pa2)
        assert set(coords.coord_map) == expected
        assert set(attrs.coords) == expected

    def test_extra_attrs(self, random_patch):
        """Ensure extra attributes are added to patch."""
        patch = random_patch.update_attrs(new_attr=10)
        coords, attrs = merge_compatible_coords_attrs(patch, random_patch)
        assert attrs.get("new_attr") == 10

    def test_different_dims(self, random_patch):
        """Ensure we can merge patches which share some dims but not all."""
        patch1 = random_patch
        patch2 = random_patch.rename_coords(time="money")
        coord, attrs = merge_compatible_coords_attrs(
            patch1, patch2, dim_intersection=True
        )
        assert set(coord.dims) == (set(patch1.dims) | set(patch2.dims))


class TestConcatenate:
    """Tests for concatenating spools."""

    def test_different_dims_raises(self, random_patch):
        """Patches can't be concated when they have different dims."""
        p1 = random_patch
        p2 = random_patch.rename_coords(time="money")
        msg = "Cannot concatenate"
        with pytest.raises(PatchCoordinateError, match=msg):
            concatenate_patches([p1, p2], time=None)

    def test_duplicate_patches_existing_dim(self, random_patch):
        """Ensure duplicate patches are concatenated together."""
        spool = dc.spool([random_patch, random_patch])
        out = concatenate_patches(spool, time=None)
        assert len(out) == 1
        patch = out[0]
        time_coord = patch.get_coord("time")
        old_coord = random_patch.get_coord("time")
        assert len(time_coord) == 2 * len(old_coord)
        # Also ensure the new coord is just the old one repeated twice.
        val1, val2 = patch.get_array("time"), random_patch.get_array("time")
        assert np.all(val1[: len(val2)] == val2)
        assert np.all(val1[len(val2) :] == val2)

    def test_different_lens(self, random_patch):
        """Ensure different lengths can be chunked on same dim."""
        patches = [random_patch] * 6
        out = concatenate_patches(patches, time=2)
        assert len(out) == 3

    def test_new_dim(self, random_patch):
        """Ensure we create a new dimension."""
        patches = [random_patch, random_patch]
        out = concatenate_patches(patches, zoolou=None)
        assert len(out) == 1
        patch = out[0]
        assert "zoolou" in patch.dims
        coord = patch.get_coord("zoolou")
        assert len(coord) == len(patches)

    def test_concat_chunk_to_new_dimension(self, random_patch):
        """Ensure the new dimension can be chunked by an int value."""
        # When new_dim = 1 it should only add a new dimension to each patch
        # and not change the original shape.
        spool = dc.spool([random_patch] * 6)
        # Test for single values along new dimension
        new = spool.concatenate(new_dim=1)
        assert len(new) == len(spool)
        for patch in new:
            coord = patch.get_coord("new_dim")
            assert len(coord) == 1
        # Test for concatenating two patches together
        new = spool.concatenate(new_dim=2)
        assert len(new) == len(spool) // 2
        for patch in new:
            coord = patch.get_coord("new_dim")
            assert len(coord) == 2

    def test_spool_up(self, random_patch):
        """Ensure a patch is returned if the wrapper is used."""
        func = _spool_up(concatenate_patches)
        out = func([random_patch] * 3, time=None)
        assert isinstance(out, dc.BaseSpool)

    def test_new_dim_spool(self, random_patch):
        """Ensure a patch with new dim can be retrieved from spool."""
        spool = dc.spool([random_patch, random_patch])
        spool_concat = spool.concatenate(wave_rank=None)
        assert len(spool_concat) == 1
        patch = spool_concat[0]
        assert "wave_rank" in patch.dims
        assert len(patch.get_coord("wave_rank")) == len(spool)

    def test_patch_with_gap(self, random_patch):
        """Ensure a patch with a time gap still concats."""
        # Create a spool with patches that have a large gap
        time = random_patch.get_coord("time")
        one_hour = dc.to_timedelta64(3600)
        patch2 = random_patch.update_coords(time_min=time.max() + one_hour)
        spool = dc.spool([random_patch, patch2])
        # chunk rightfully wouldn't merge these patches, but concatenate will.
        merged = spool.concatenate(time=None)
        assert len(merged) == 1
        assert isinstance(merged[0], dc.Patch)

    def test_bad_kwargs(self, random_patch):
        """Ensure bad number of keywords raise."""
        msg = "Exactly one keyword argument"
        with pytest.raises(ParameterError, match=msg):
            concatenate_patches([random_patch])
        with pytest.raises(ParameterError, match=msg):
            concatenate_patches([random_patch], time=None, distance=None)

    def test_concat_along_non_dim(self, random_patch):
        """Ensure we can concat along non dims."""
        patches = [random_patch.mean("time") for _ in range(10)]
        out = concatenate_patches(patches, time=None)
        assert len(out) == 1
        patch = out[0]
        coord = patch.get_coord("time")
        assert len(coord) == 10

    def test_concat_different_sizes(self, random_patch):
        """Ensure coordinates with different sizes (along concat axis) work."""
        p1 = random_patch.select(time=(0, 10), samples=True)
        p2 = random_patch.select(time=(15, 20), samples=True)
        out = concatenate_patches([p1, p2], time=None)
        assert len(out) == 1
        patch = out[0]
        new_time = patch.get_array("time")
        old_times = np.concatenate([p1.get_array("time"), p2.get_array("time")])
        assert np.all(new_time == old_times)

    def test_concat_normal_with_non_dim(self, spool_with_non_coords):
        """Ensure normal and non-dim patches can be concatenated together."""
        old_arrays = [x.get_array("time") for x in spool_with_non_coords]
        old_array = np.concatenate(old_arrays)
        out = spool_with_non_coords.concatenate(time=None)
        assert len(out) == 1
        patch = out[0]
        new_array = patch.get_array("time")
        # The coordinates should be the same length
        assert sum([len(x) for x in old_arrays]) == len(new_array)
        # The array values should either both be NaN or nearly equal
        both_nan = pd.isnull(old_array) & pd.isnull(new_array)
        try:
            nearly_eq = np.isclose(old_array, new_array)
        except TypeError:
            nearly_eq = old_array == new_array
        assert np.all(both_nan | nearly_eq)

    def test_private_coords_dropped(self, random_patch):
        """Ensure private coords don't interfere with concat along new dim."""
        pa1 = random_patch.update_coords(_private_1=(None, np.array([1, 2, 3])))
        pa2 = random_patch.update_coords(_private_1=(None, np.array([2, 2, 2])))
        spool = dc.spool([pa1, pa2])
        out = spool.concatenate(time_new=None)
        assert len(out) == 1
        assert out[0].shape[-1] == 2

    def test_concat_dropped_coord(self, random_spool):
        """Ensure patches after dropping a coordinate can be concatenated together
        and the concatenated patch can have a new dimension.
        """
        sp = random_spool
        pa_list = []
        for pa in sp:
            pa_dft = pa.dft("time")
            cm = pa_dft.coords
            pa_dft_dropped_time = pa_dft.update(coords=cm.update(time=None))
            pa_list.append(pa_dft_dropped_time)
        sp_dft = dc.spool(pa_list)
        time_min_coord = sp_dft.get_contents()["time_min"]
        sp_concat = sp_dft.concatenate(time_min=None)
        pa_concat = sp_concat[0]
        updated_coords = pa_concat.coords.update(time_min=time_min_coord)
        pa_concat = pa_concat.update(coords=updated_coords)
        assert pa_concat.shape[-1] == len(sp)
        assert "time_min" in pa_concat.dims


class TestStackPatches:
    """Tests for stacking (adding) spool content."""

    def test_stack_data(self):
        """
        Try the stack method on an example spool that has repeated
        copies of the same patch with different times. Check data.
        """
        # Grab the example spool which has repeats of the same patch
        # but with different time dimensions. Stack the patches.
        spool = dc.get_example_spool()
        stack_patch = stack_patches(spool, dim_vary="time")
        # We expect the sum/stack to be same as a multiple of
        # the first patch's data.
        baseline = float(len(spool)) * spool[0].data
        assert np.allclose(baseline, stack_patch.data)

    def test_same_dim_different_shape(self, random_spool):
        """Ensure when stack dimensions have different shape an error is raised."""
        # Create a spool with two patches, each with time dim but with different
        # lengths.
        patch1, patch2 = random_spool[:2]
        patch2 = patch2.select(time=(1, 30), samples=True)
        spool = dc.spool([patch1, patch2])
        # Check that warnings/exceptions are raised.
        msg = "Patches are not compatible"
        with pytest.raises(IncompatiblePatchError, match=msg):
            stack_patches(spool, dim_vary="time", check_behavior="raise")
        # Or a warning issued.
        with pytest.warns(UserWarning, match=msg):
            stack_patches(spool, dim_vary="time", check_behavior="warn")

    def test_different_dimensions(self, random_spool):
        """Tests for when the spool has patches with different dimensions."""
        new_patch = random_spool[0].rename_coords(time="money")
        spool = dc.spool([random_spool[1], new_patch])
        msg = "not compatible for merging"
        with pytest.warns(UserWarning, match=msg):
            stack_patches(spool, dim_vary="time", check_behavior="warn")

    def test_bad_dim_vary(self, random_spool):
        """Ensure when dim_vary is not in patch an error is raised."""
        with pytest.raises(PatchCoordinateError):
            stack_patches(random_spool, dim_vary="money")

    def test_stack_coords(self):
        """
        Try the stack method on an example spool that has repeated
        copies of the same patch with different times. Check coords.
        """
        # Grab the example spool which has repeats of the same patch
        # but with different time dimensions. Stack the patches.
        spool = dc.get_example_spool()
        stack_patch = stack_patches(spool, dim_vary="time")
        # check that 'time' coordinates has same step as original patch
        time_coords = stack_patch.coords.coord_map["time"]
        orig_time_coords = spool[0].coords.coord_map["time"]
        assert time_coords.step == orig_time_coords.step
        # check that distance coordinates are the same
        dist_coords = stack_patch.coords.coord_map["distance"]
        orig_dist_coords = spool[0].coords.coord_map["distance"]
        assert dist_coords.start == orig_dist_coords.start
        assert dist_coords.stop == orig_dist_coords.stop
        assert dist_coords.step == orig_dist_coords.step
        assert dist_coords.units == orig_dist_coords.units


class TestGetPatchName:
    """Tests for getting the default name from patch sources."""

    def test_single_patch(self, random_patch):
        """Ensure the random patch can have a name."""
        out = get_patch_names(random_patch)
        assert isinstance(out, pd.Series)
        assert len(out) == 1

    def test_spool(self, random_spool):
        """Ensure names can be generated from a spool as well."""
        out = get_patch_names(random_spool)
        assert len(out) == len(random_spool)

    def test_empty(self):
        """Ensure an empty thing returns a series of the right type."""
        out = get_patch_names([])
        assert isinstance(out, pd.Series)

    def test_name_column_exists(self, random_spool):
        """If the name or path field already exist this should be used."""
        df = random_spool.get_contents().assign(name=lambda x: np.arange(len(x)))
        # This should convert to string type and return.
        names = get_patch_names(df)
        assert np.all(names.values == np.arange(len(df)).astype(str))

    def test_path_column(self, random_directory_spool):
        """Ensure the path column works."""
        names = get_patch_names(random_directory_spool)
        df = random_directory_spool.get_contents()
        expected = pd.Series([x[-1].split(".")[0] for x in df["path"].str.split("/")])
        assert np.all(names == expected)

    def test_path_column_leave_extension(self, random_directory_spool):
        """If extension is false the file extension should remain."""
        names = get_patch_names(random_directory_spool, strip_extension=False)
        assert "." in names.iloc[0]



"""A spectrogram visualization."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

import matplotlib.pyplot as plt

from dascore.constants import PatchType
from dascore.utils.patch import patch_function


def _get_other_dim(dim, dims):
    if not isinstance(dim, str):
        raise TypeError(f"Expected 'dim' to be a string, got {type(dim).__name__}.")
    if dim not in dims:
        raise ValueError(f"The dimension '{dim}' is not in patch's dimensions {dims}.")
    if len(dims) == 1:
        return None
    else:
        return dims[0] if dims[1] == dim else dims[1]


@patch_function()
def spectrogram(
    patch: PatchType,
    ax: plt.Axes | None = None,
    dim="time",
    aggr_domain="frequency",
    cmap="bwr",
    scale: float | Sequence[float] | None = None,
    scale_type: Literal["relative", "absolute"] = "relative",
    log=False,
    show=False,
    **kwargs,
) -> plt.Axes:
    """
    Plot a spectrogram of a patch.

    Parameters
    ----------
    patch : PatchType
        The Patch object.
    ax : matplotlib.axes.Axes or None, optional
        A matplotlib axis object. If None, creates a new axis.
    dim : str, optional
        Dimension along which the spectrogram is being plotted.
        Default is "time".
    aggr_domain : str, optional
        "time" or "frequency" in which the mean value of the other
        dimension is calculated. No need to specify if the other
        dimension's coordinate size is 1. Default is "frequency".
    cmap : str or matplotlib.colors.Colormap, optional
        A matplotlib colormap string or instance. Set to None to not plot the
        colorbar. Default is "bwr".
    scale : float, tuple of floats, or None, optional
        If not None, controls the saturation level of the colorbar.
        Values can be a single float or a length-2 tuple specifying upper
        and lower limits. See `scale_type` for more details.
    scale_type : {"relative", "absolute"}, optional
        Specifies the type of scaling:
            - "relative": Scale based on half the dynamic range in the patch.
            - "absolute": Scale based on absolute values provided to `scale`.
        Default is "relative".
    log : bool, optional
        If True, visualize the common logarithm of the absolute values of patch data.
    show : bool, optional
        If True, show the plot. Otherwise, just return the axis.
    **kwargs : dict, optional
        Passed to `scipy.signal.spectrogram` to control spectrogram options.
        See its documentation for options.
    """
    dims = patch.dims
    if len(dims) > 2 or len(dims) < 1:
        raise ValueError("Can only make spectrogram of 1D or 2D patches.")

    other_dim = _get_other_dim(dim, dims)
    if other_dim is not None:
        if aggr_domain == "time":
            patch_aggr = patch.aggregate(other_dim, method="mean", dim_reduce="squeeze")
            spec = patch_aggr.spectrogram(dim)
        elif aggr_domain == "frequency":
            _spec = patch.spectrogram(dim).squeeze()
            spec = _spec.aggregate(other_dim, method="mean").squeeze()
        else:
            raise ValueError(
                f"The aggr_domain '{aggr_domain}' should be 'time' or 'frequency'."
            )
    else:
        spec = patch.spectrogram(dim, **kwargs)
    return spec.viz.waterfall(
        ax=ax, cmap=cmap, scale=scale, scale_type=scale_type, log=log, show=show
    )



"""
Common tests for IO operations.

The objective of this test file is to reduce the amount of repetition
in testing common operations with various formats. Format specific
tests should go in their respective test modules. Tests for *how* specific
IO functions (i.e., not that they work on various files) should go in
test_io_core.py
"""

from __future__ import annotations

from contextlib import contextmanager, suppress
from functools import cache
from io import BytesIO
from operator import eq, ge, le
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.exceptions import MissingOptionalDependencyError
from dascore.io import BinaryReader
from dascore.io.ap_sensing import APSensingV10
from dascore.io.dasdae import DASDAEV1
from dascore.io.dashdf5 import DASHDF5
from dascore.io.febus import Febus2
from dascore.io.gdr import GDR_V1
from dascore.io.h5simple import H5Simple
from dascore.io.neubrex import NeubrexDASV1, NeubrexRFSV1
from dascore.io.optodas import OptoDASV8
from dascore.io.pickle import PickleIO
from dascore.io.prodml import ProdMLV2_0, ProdMLV2_1
from dascore.io.segy import SegyV1_0
from dascore.io.sentek import SentekV5
from dascore.io.silixah5 import SilixaH5V1
from dascore.io.sintela_binary import SintelaBinaryV3
from dascore.io.tdms import TDMSFormatterV4713
from dascore.io.terra15 import (
    Terra15FormatterV4,
    Terra15FormatterV5,
    Terra15FormatterV6,
)
from dascore.utils.downloader import fetch, get_registry_df
from dascore.utils.misc import all_close, iterate

# --- Fixtures

# These fixtures are for attaching FiberIO objects to a suite of common
# read, scan, and get_format tests that all FiberIO instances which implement
# more than just a write method should pass.
# The format is {FiberIO class: [path_or_fetch_name]} where fetch_name
# is the name passed to dascore.utils.downloader.fetch to load the file.
# See the docs on adding a new IO format, in the contributing section,
# for more details.
COMMON_IO_READ_TESTS = {
    SintelaBinaryV3(): ("sintela_binary_v3_test_1.raw",),
    GDR_V1(): ("gdr_1.h5",),
    NeubrexDASV1(): ("neubrex_das_1.h5",),
    NeubrexRFSV1(): ("neubrex_dss_forge.h5", "neubrex_dts_forge.h5"),
    SilixaH5V1(): ("silixa_h5_1.hdf5",),
    ProdMLV2_0(): ("prodml_2.0.h5", "opta_sense_quantx_v2.h5"),
    ProdMLV2_1(): (
        "prodml_2.1.h5",
        "iDAS005_hdf5_example.626.h5",
    ),
    H5Simple(): ("h5_simple_2.h5", "h5_simple_1.h5"),
    DASDAEV1(): ("example_dasdae_event_1.h5",),
    APSensingV10(): ("ap_sensing_1.hdf5",),
    Febus2(): ("febus_1.h5",),
    OptoDASV8(): ("opto_das_1.hdf5",),
    DASDAEV1(): ("example_dasdae_event_1.h5",),
    TDMSFormatterV4713(): ("sample_tdms_file_v4713.tdms",),
    Terra15FormatterV4(): (
        "terra15_das_1_trimmed.hdf5",
        "terra15_das_unfinished.hdf5",
    ),
    Terra15FormatterV5(): ("terra15_v5_test_file.hdf5",),
    Terra15FormatterV6(): ("terra15_v6_test_file.hdf5",),
    Terra15FormatterV6(): ("terra15_v6_test_file.hdf5",),
    SegyV1_0(): ("conoco_segy_1.sgy",),
    DASHDF5(): ("PoroTomo_iDAS_1.h5",),
    SentekV5(): ("DASDMSShot00_20230328155653619.das",),
}

# This tuple is for fiber io which support a write method and can write
# generic patches. If the patch has to be in some special form, for example
# only flat patches can be written to WAV, don't put it here.
COMMON_IO_WRITE_TESTS = (
    PickleIO(),
    DASDAEV1(),
)

# Specifies data registry entries which should not be tested.
SKIP_DATA_FILES = {"whale_1.hdf5", "brady_hs_DAS_DTS_coords.csv"}


@cache
def _cached_read(path, io=None):
    """
    A function to read data without any params and cache.
    This ensures each files is read at most twice.
    """
    if io is None:
        read = dc.read
    else:
        read = io.read
    with skip_missing_dependency():
        out = read(path)
    return out


def _get_flat_io_test():
    """Flatten list to [(fiberio, path)] so it can be parametrized."""
    flat_io = []
    for io, fetch_name_list in COMMON_IO_READ_TESTS.items():
        for fetch_name in iterate(fetch_name_list):
            flat_io.append([io, fetch_name])
    return flat_io


@contextmanager
def skip_missing_dependency():
    """Skip if missing dependencies found."""
    try:
        yield
    except MissingOptionalDependencyError:
        pytest.skip("Missing optional dep to read file.")


@pytest.fixture(scope="session", params=list(COMMON_IO_READ_TESTS))
def io_instance(request):
    """Fixture for returning fiber io instances."""
    return request.param


@pytest.fixture(scope="session", params=_get_flat_io_test())
def io_path_tuple(request):
    """
    A fixture which returns io instance, path_to_file.
    This is used for common testing.
    """
    io, fetch_name = request.param
    return io, fetch(fetch_name)


@pytest.fixture(scope="session", params=get_registry_df()["name"])
def data_file_path(request):
    """A fixture of all data files. Will download if needed."""
    param = request.param
    # Some files should be skipped if not DAS or too big.
    if str(param) in SKIP_DATA_FILES:
        pytest.skip(f"Skipping {param}")
    return fetch(request.param)


@pytest.fixture(scope="session")
def read_spool(data_file_path):
    """Read each file into a spool."""
    with skip_missing_dependency():
        out = dc.read(data_file_path)
    return out


@pytest.fixture(scope="session")
def scanned_attrs(data_file_path):
    """Read each file into a spool."""
    with skip_missing_dependency():
        out = dc.scan(data_file_path)
    return out


@pytest.fixture(scope="session", params=COMMON_IO_WRITE_TESTS)
def fiber_io_writer(request):
    """Meta fixture for IO that implement write."""
    return request.param


# --- Helper functions


def _assert_coords_attrs_match(patch):
    """Ensure both the coordinates and attributes match on patch."""
    attrs = patch.attrs
    coords = patch.coords
    assert attrs.dim_tuple == coords.dims
    for dim in attrs.dim_tuple:
        coord = patch.get_coord(dim)
        assert coord.min() == getattr(attrs, f"{dim}_min")
        assert coord.max() == getattr(attrs, f"{dim}_max")
        assert coord.step == getattr(attrs, f"{dim}_step")


def _assert_op_or_close(val1, val2, op):
    """Assert op(val1, val2) or isclose(val1, val2)."""
    meets_eq = op(val1, val2)
    both_null = pd.isnull(val1) and pd.isnull(val2)
    all_close_vals = all_close(val1, val2)
    if meets_eq or both_null or all_close_vals:
        return
    msg = f"{val1} and {val2} do not pass {op} or all close."
    raise AssertionError(msg)


# --- Tests

DIM_RELATED_ATTRS = ("{dim}_min", "{dim}_max", "{dim}_step")


class TestGetFormat:
    """Test suite for getting the version/format of files."""

    def test_expected_version(self, io_path_tuple):
        """Each io should get its own version/name for its test files."""
        io, path = io_path_tuple
        expected = (io.name, io.version)
        out = io.get_format(path)
        assert out == expected

    def test_each_file_has_version(self, data_file_path):
        """We should be able to get each file's format/version."""
        out = dc.get_format(data_file_path)
        assert out
        assert len(out) == 2
        format_str, version_str = out
        assert isinstance(format_str, str)
        assert isinstance(version_str, str)

    def test_random_textfile_isnt_format(self, io_instance, dummy_text_file):
        """Ensure a dummy text file the format (it isn't any fiber format)."""
        assert not io_instance.get_format(dummy_text_file)

    def test_random_h5_isnt_format(self, io_instance, generic_hdf5):
        """Ensure a dummy h5 file the format (it isn't any fiber format)."""
        assert not io_instance.get_format(generic_hdf5)

    def test_all_other_files_arent_format(self, io_instance):
        """All other data files should not show up as this format."""
        for other_io, data_files in COMMON_IO_READ_TESTS.items():
            if isinstance(other_io, type(io_instance)):
                continue
            for key in data_files:
                path = fetch(key)
                out = io_instance.get_format(path)
                if out:
                    format_name, version = out
                    assert version != io_instance.version


class TestRead:
    """Test suite for reading formats."""

    def test_read_returns_spools(self, io_path_tuple):
        """Each read function must return a spool."""
        io, path = io_path_tuple
        out = _cached_read(path, io=io)
        assert isinstance(out, dc.BaseSpool)
        assert all([isinstance(x, dc.Patch) for x in out])

    def test_coord_attrs_match(self, read_spool):
        """The attributes and coordinate should match in min/max/step."""
        for patch in read_spool:
            _assert_coords_attrs_match(patch)

    def test_time_coords_are_datetimes(self, read_spool):
        """Ensure the time coordinates have the type of datetime."""
        for patch in read_spool:
            with suppress(KeyError):
                time = patch.get_coord("time")
                assert "datetime64" in str(np.dtype(time.dtype))

    def test_read_stream(self, io_path_tuple):
        """If the format supports reading from a stream, test it out."""
        io, path = io_path_tuple
        req_type = getattr(io.read, "_required_type", None)
        if not isinstance(req_type, BinaryReader):
            msg = f"{io} doesn't require a read type."
            pytest.skip(msg)
        spool1 = _cached_read(path)
        # write file contents to bytes io and ensure it can be read.
        bio = BytesIO()
        bio.write(Path(io_path_tuple).read_bytes())
        bio.seek(0)
        spool2 = io.read(bio)
        for patch1, patch2 in zip(spool1, spool2):
            assert patch1.equals(patch2)

    def test_slice_single_dim_both_ends(self, io_path_tuple):
        """
        Ensure each dimension can be passed as an argument to `read` and
        a patch containing the requested data is returned.
        """
        io, path = io_path_tuple
        with skip_missing_dependency():
            attrs_from_file = dc.scan(path)
        assert len(attrs_from_file)
        # skip files that have more than one patch for now
        # TODO just write better test logic to handle this case.
        if len(attrs_from_file) > 1:
            pytest.skip("Haven't implemented test for multipatch files.")
        attrs_init = attrs_from_file[0]
        for dim in attrs_init.dim_tuple:
            start = getattr(attrs_init, f"{dim}_min")
            stop = getattr(attrs_init, f"{dim}_max")
            duration = stop - start
            # first test double ended query
            trim_tuple = (start + duration / 10, start + 2 * duration / 10)
            spool = io.read(path, **{dim: trim_tuple})
            assert len(spool) == 1
            patch = spool[0]
            _assert_coords_attrs_match(patch)
            coord = patch.get_coord(dim)
            _assert_op_or_close(coord.min(), trim_tuple[0], ge)
            _assert_op_or_close(coord.max(), trim_tuple[1], le)
            # then single-ended query on start side
            trim_tuple = (start + duration / 10, ...)
            spool = io.read(path, **{dim: trim_tuple})
            assert len(spool) == 1
            patch = spool[0]
            attrs = patch.attrs
            _assert_coords_attrs_match(patch)
            coord = patch.get_coord(dim)
            _assert_op_or_close(coord.min(), trim_tuple[0], ge)
            _assert_op_or_close(coord.max(), getattr(attrs, f"{dim}_max"), eq)
            # then single-ended query on end side
            trim_tuple = (None, start + duration / 10)
            spool = io.read(path, **{dim: trim_tuple})
            assert len(spool) == 1
            patch = spool[0]
            attrs = patch.attrs
            _assert_coords_attrs_match(patch)
            coord = patch.get_coord(dim)
            _assert_op_or_close(coord.min(), getattr(attrs, f"{dim}_min"), eq)
            _assert_op_or_close(coord.max(), trim_tuple[1], le)


class TestScan:
    """Tests for generic scanning."""

    def test_scan_basics(self, data_file_path):
        """Ensure each file can be scanned."""
        with skip_missing_dependency():
            attrs_list = dc.scan(data_file_path)
        assert len(attrs_list)

        for attrs in attrs_list:
            assert isinstance(attrs, dc.PatchAttrs)
            assert str(attrs.path) == str(data_file_path)

    def test_scan_has_version_and_format(self, io_path_tuple):
        """Scan output should contain version and format."""
        io, path = io_path_tuple
        with skip_missing_dependency():
            attr_list = io.scan(path)
        for attrs in attr_list:
            assert attrs.file_format == io.name
            assert attrs.file_version == io.version

    def test_time_coord_is_time(self, scanned_attrs):
        """Ensure scanned attrs have correct dtype for time."""
        for patch_attr in scanned_attrs:
            with suppress(KeyError):
                time = patch_attr.coords["time"]
                assert "datetime64" in str(np.dtype(time.dtype))

    def test_dist_coord_is_float_or_int(self, scanned_attrs):
        """Distance can be either float or int, but must be numeric."""
        for patch_attr in scanned_attrs:
            with suppress(KeyError):
                distance = patch_attr.coords["distance"]
                dtype = np.dtype(distance.dtype)
                assert np.issubdtype(dtype, np.number)

    def test_no_bytes(self, scanned_attrs):
        """Sometimes bytes are returned from scanning, we need str."""
        for patch_attr in scanned_attrs:
            model = patch_attr.model_dump()
            for key, value in model.items():
                assert not isinstance(value, bytes | np.bytes_)


class TestWrite:
    """Tests for writing data to disk."""

    @pytest.fixture(scope="session")
    def written_fiber_path(self, random_patch, fiber_io_writer, tmp_path_factory):
        """Write patch to disk, return path."""
        tmp_path = Path(tmp_path_factory.mktemp("generic_write_test"))
        pre_ext = list(iterate(fiber_io_writer.preferred_extensions))
        ext = "" if not len(pre_ext) else pre_ext[0]
        path = tmp_path / f"{fiber_io_writer.name}.{ext}"
        fiber_io_writer.write(random_patch, path)
        return path

    def test_write_random_patch(self, written_fiber_path):
        """Ensure the random patch can be written."""
        assert written_fiber_path.exists()

    def test_roundtrip(self, random_patch, written_fiber_path, fiber_io_writer):
        """If the writer can read, ensure round-tripping patch is equal."""
        if not fiber_io_writer.implements_read:
            pytest.skip("FiberIO doesn't implement read")
        new = fiber_io_writer.read(written_fiber_path)[0]
        assert new == random_patch


class TestIntegration:
    """Test suite for generic scanning."""

    def test_scan_attrs_match_patch_attrs(self, data_file_path):
        """We need to make sure scan and patch attrs are identical."""
        # Since dasdae format stores attrs and coords, we need to
        # skip events created before coords/attrs were more closely
        # aligned.
        if data_file_path.name == "example_dasdae_event_1.h5":
            return
        comp_attrs = (
            "data_type",
            "data_units",
            "tag",
            "network",
        )
        with skip_missing_dependency():
            scan_attrs_list = dc.scan(data_file_path)
        patch_attrs_list = [x.attrs for x in _cached_read(data_file_path)]
        assert len(scan_attrs_list) == len(patch_attrs_list)
        for pat_attrs1, scan_attrs2 in zip(patch_attrs_list, scan_attrs_list):
            assert pat_attrs1.dims == scan_attrs2.dims
            # first compare dimensions are related attributes
            for dim in pat_attrs1.dim_tuple:
                assert getattr(pat_attrs1, f"{dim}_min") == getattr(
                    pat_attrs1, f"{dim}_min"
                )
                for dim_attr in DIM_RELATED_ATTRS:
                    attr_name = dim_attr.format(dim=dim)
                    attr1 = getattr(pat_attrs1, attr_name)
                    attr2 = getattr(pat_attrs1, attr_name)
                    assert attr1 == attr2
            # then other expected attributes.
            for attr_name in comp_attrs:
                patch_attr = getattr(pat_attrs1, attr_name)
                scan_attr = getattr(scan_attrs2, attr_name)
                assert scan_attr == patch_attr



"""
A few mappings that might be useful.

We can't simply use types.MappingProxyType because it can't be pickled.
"""

from __future__ import annotations

from collections.abc import Mapping as ABCMap


class FrozenDict(ABCMap):
    """
    An immutable wrapper around dictionaries that implements the complete
    :py:class:`collections.Mapping` interface. It can be used as a drop-in
    replacement for dictionaries where immutability is desired.

    Notes
    -----
    This implementation was Inspired by the no-longer maintained package
    frozen-dict (https://github.com/slezica/python-frozendict)

    By design, changes in the original dict are not reflected in the frozen
    dict so that the hash doesn't break.
    """

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def new(self, **kwargs):
        """Copy the contents  and update with new values."""
        contents = dict(self._dict)
        contents.update(kwargs)
        return self.__class__(**contents)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._dict!r}>"

    def _hash_contents(self):
        """Returns a hash of the dictionary."""
        out = 0
        for key, value in self._dict.items():
            out ^= hash((key, value))
        return out

    def __hash__(self):
        if self._hash is None:
            self._hash = self._hash_contents()
        return self._hash



"""
Tests for sintela binary format.
"""

import shutil
from pathlib import Path

import pytest

from dascore.exceptions import InvalidFiberFileError
from dascore.io.sintela_binary import SintelaBinaryV3
from dascore.utils.downloader import fetch


class TestScanSintelaBinary:
    """Tests for scanning a binary file."""

    @pytest.fixture(scope="class")
    def extra_bytes_file(self, tmp_path_factory):
        """Create a sintela binary file with extra bytes."""
        tmp = tmp_path_factory.mktemp("sintela_binary")
        binary_path = Path(fetch("sintela_binary_v3_test_1.raw"))
        new = tmp / "extra_bytes.raw"
        shutil.copy(binary_path, new)

        with open(new, "ab") as fi:
            fi.write(b"some_bytes_des_is")

        return new

    def test_extra_bytes_raises(self, extra_bytes_file):
        """Ensure a file with extra bytes raises an exception."""
        fiber_io = SintelaBinaryV3()
        with pytest.raises(InvalidFiberFileError):
            fiber_io.scan(extra_bytes_file)



"""
Support for OptoDAS format.

This is used by the OptoDAS interrogator made by Alcatel Submarine Networks.

More info here: https://web.asn.com/
"""
from __future__ import annotations
from .core import OptoDASV8, OptoDASV10



"""
Support for Febus format.

This is used by the Febus DAS interrogator.

More info about febus can be found here: https://www.febus-optics.com/en/
"""
from __future__ import annotations
from .core import Febus2



"""Implements logic to apply formatting changes to patches from transformations."""

from __future__ import annotations

import abc

from dascore.utils.misc import iterate


class BaseTransformatter(abc.ABC):
    """Base model for helping to apply transformation format changes."""

    forward_prefix: str = ""
    inverse_prefix: str = ""

    def _forward_rename(self, name):
        """Rename the dimension for forward transform."""
        if name.startswith(self.inverse_prefix):
            return name[len(self.inverse_prefix) :]
        return f"{self.forward_prefix}{name}"

    def _inverse_rename(self, name):
        """Rename the dimension for backward transform."""
        if name.startswith(self.forward_prefix):
            return name[len(self.forward_prefix) :]
        return f"{self.inverse_prefix}{name}"

    def rename_dims(self, dims, index=None, forward=True):
        """Rename the dimensions."""
        func = self._forward_rename if forward else self._inverse_rename
        new = list(iterate(dims))
        index_list = iterate(index) if index is not None else range(len(new))
        for index in index_list:
            new[index] = func(new[index])
        return tuple(new)


class FourierTransformatter(BaseTransformatter):
    """Formatters."""

    forward_prefix: str = "ft_"
    inverse_prefix: str = "ift_"



"""Tests for the spectrogram transformation."""

from __future__ import annotations

import pytest

import dascore as dc
from dascore.transform.spectro import spectrogram


class TestSpectroTransform:
    """Tests for transforming regular patches into spectrograms."""

    @pytest.fixture()
    def spec_patch(self, random_patch):
        """Simple patch trasnformed to spectrogram."""
        patch = random_patch.set_units("m/s")
        return patch.spectrogram("time")

    def test_units(self, random_patch):
        """Ensure units were properly converted."""
        patch = random_patch.set_units("m/s")
        spec_patch = patch.spectrogram("time")
        # first check coord units
        coord1 = patch.get_coord("time")
        coord2 = spec_patch.get_coord("ft_time")
        units1 = dc.get_quantity(coord1.units)
        units2 = dc.get_quantity(coord2.units)
        assert units1 == 1 / units2
        # then check data units
        data_units1 = dc.get_quantity(patch.attrs.data_units)
        data_units2 = dc.get_quantity(spec_patch.attrs.data_units)
        assert data_units1 * units1 == data_units2

    def test_spec_patch_dimensions(self, spec_patch, random_patch):
        """Ensure expected dimensions now exist."""
        dims = spec_patch.dims
        # dims should have been added
        assert len(dims) > len(random_patch.dims)
        assert set(dims) == (set(random_patch.dims) | {"ft_time"})

    def test_transformed_coord(self, spec_patch, random_patch):
        """
        The start values of transformed dimension should be comparable and the
        units unchanged.
        """
        time_1 = random_patch.get_coord("time")
        time_2 = spec_patch.get_coord("time")
        assert time_1.units == time_2.units

    def test_time_first(self, random_patch):
        """Ensure the spectrogram still works when time dim is first."""
        transposed = random_patch.transpose(*("time", "distance"))
        out = spectrogram(transposed, dim="time")
        assert isinstance(out, dc.Patch)

    def test_transpose(self, random_patch):
        """Ensure when patch dim order is different it still works."""
        out = random_patch.transpose().spectrogram("time")
        assert set(out.dims) == (set(random_patch.dims) | {"ft_time"})

    def test_distance_dim(self, random_patch):
        """Ensure distance dimension works."""
        out = random_patch.transpose().spectrogram("distance")
        assert set(out.dims) == (set(random_patch.dims) | {"ft_distance"})

    def test_time_range_unchanged(self, dispersion_patch):
        """Ensure time axis isn't outside original bounds, see #286."""
        spectro = dispersion_patch.spectrogram(dim="time")
        # the new time on the spectrogram should be contained in the original
        original_time = dispersion_patch.get_coord("time")
        new_time = spectro.get_coord("time")
        assert new_time.min() >= original_time.min()
        assert new_time.max() <= original_time.max()



"""
Support for Neubrex H5 DSS/DTS files.

This module was written to read the DSS/DTS files created by Neubrex for
the Forge dataset: https://gdr.openei.org/submissions/1565

The citation for the dataset is:

Energy and Geoscience Institute at the University of Utah. (2023).
Utah FORGE: Well 16B(78)-32 2023 Neubrex Energy Services Circulation
Test Period with Fiber Optics Monitoring [data set].
Retrieved from https://dx.doi.org/10.15121/2222469.
"""
from __future__ import annotations

from .core import NeubrexRFSV1, NeubrexDASV1



"""
Tests for applying just in time compilations.
"""

from functools import cache

import numpy as np
import pytest

from dascore.utils.jit import maybe_numba_jit
from dascore.utils.misc import optional_import, suppress_warnings


class TestMaybeNumbaJit:
    """Tests for optional jit'ing."""

    def test_jit(self):
        """Ensure the jit works. Only test if numba installed."""
        pytest.importorskip("numba")

        @maybe_numba_jit
        def my_jit(ar):
            return ar

        ar = np.array([1, 2, 3])
        new_ar = my_jit(ar)
        assert np.all(new_ar == ar)

    def test_warning(self):
        """When numba is not installed ensure a warning is issued."""

        @maybe_numba_jit(_missing_numba=True)
        def _jit_test_func(ar):
            return ar

        match = "can be compiled to improve performance"
        with pytest.warns(UserWarning, match=match):
            _jit_test_func(np.array([1, 2, 3]))

    def test_raises(self):
        """Ensure an error is raised when specified by the decorator."""

        @maybe_numba_jit(required=True, _missing_numba=True)
        def _jit_test_func(ar):
            return ar

        match = "requires python module"
        with pytest.raises(ImportError, match=match):
            _jit_test_func(np.array([1, 2, 3]))

    def test_example(self):
        """Test docstring examples."""
        pytest.importorskip("numba")

        @maybe_numba_jit(nopython=True, nogil=True)
        def _jit_func(array):
            return array

        @cache
        def jit_wrapper():
            numba = optional_import("numba")

            @maybe_numba_jit
            def jit_func(array):
                for a in numba.prange(len(array)):
                    pass
                return array

            return jit_func

        out = jit_wrapper()(np.array([1, 2, 3]))
        assert isinstance(out, np.ndarray)

    def test_numba_used_in_function(self):
        """Tests for numba used in function without being imported."""
        pytest.importorskip("numba")

        @maybe_numba_jit(nopython=True, nogil=True)
        def _my_jit(array):
            for sub in numba.prange(len(array)):  # noqa
                pass
            return array

        array = np.array([1, 2, 3])
        out = _my_jit(array)
        assert np.all(out == array)

    def test_prange_no_numba(self):
        """
        In order to make the doctests work, we had to implement a dummy
        numba module. It doesn't support everything, (barely anything) but
        should be enough for now.
        """

        @maybe_numba_jit(_missing_numba=True)
        def _my_jit(array):
            for sub in numba.prange(len(array)):  # noqa
                pass
            return array

        array = np.array([1, 2, 3])

        with suppress_warnings(UserWarning):
            out = _my_jit(array)
        assert np.all(out == array)



"""Module for spools, containers of patches."""

from __future__ import annotations

import abc
from collections.abc import Callable, Generator, Mapping, Sequence
from functools import singledispatch
from pathlib import Path
from typing import Literal, TypeVar

import numpy as np
import pandas as pd
from rich.text import Text
from typing_extensions import Self

import dascore as dc
from dascore.compat import is_array
from dascore.constants import (
    PROGRESS_LEVELS,
    WARN_LEVELS,
    ExecutorType,
    PatchType,
    attr_conflict_description,
    numeric_types,
    timeable_types,
)
from dascore.exceptions import InvalidSpoolError, ParameterError
from dascore.utils.chunk import ChunkManager
from dascore.utils.display import get_dascore_text, get_nice_text
from dascore.utils.docs import compose_docstring
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import CacheDescriptor, _spool_map
from dascore.utils.patch import (
    _force_patch_merge,
    _spool_up,
    concatenate_patches,
    get_patch_names,
    patches_to_df,
    stack_patches,
)
from dascore.utils.pd import (
    _column_or_value,
    _convert_min_max_in_kwargs,
    adjust_segments,
    filter_df,
    get_column_names_from_dim,
    get_dim_names_from_columns,
    split_df_query,
)

T = TypeVar("T")


class BaseSpool(abc.ABC):
    """Spool Abstract Base Class (ABC) for defining Spool interface."""

    _rich_style = "bold"

    @abc.abstractmethod
    def __getitem__(self, item: int | slice | np.ndarray) -> PatchType:
        """Returns a patch from the spool."""

    @abc.abstractmethod
    def __iter__(self) -> PatchType:
        """Iterate through the Patches in the spool."""

    @abc.abstractmethod
    def __len__(self) -> int:
        """Return len of spool."""

    def __rich__(self):
        """Rich rep. of spool."""
        text = get_dascore_text() + Text(" ")
        text += Text(self.__class__.__name__, style=self._rich_style)
        text += Text(" 🧵 ")
        patch_len = len(self)
        text += Text(f"({patch_len:d}")
        text += Text(" Patches)") if patch_len != 1 else Text(" Patch)")
        return text

    def __str__(self):
        return str(self.__rich__())

    __repr__ = __str__

    def __eq__(self, other) -> bool:
        """Simple equality checks on spools."""

        def _vals_equal(dict1, dict2):
            if (set1 := set(dict1)) != set(dict2):
                return False
            for key in set1:
                val1, val2 = dict1[key], dict2[key]
                if isinstance(val1, dict):
                    if not _vals_equal(val1, val2):
                        return
                # this is primarily for dataframes which have equals method.
                elif hasattr(val1, "equals"):
                    if not val1.equals(val2):
                        return False
                elif val1 != val2:
                    return False
            return True

        my_dict = self.__dict__
        other_dict = getattr(other, "__dict__", {})
        return _vals_equal(my_dict, other_dict)

    @abc.abstractmethod
    @compose_docstring(conflict_desc=attr_conflict_description)
    def chunk(
        self,
        overlap: numeric_types | timeable_types | None = None,
        keep_partial: bool = False,
        snap_coords: bool = True,
        tolerance: float = 1.5,
        conflict: Literal["drop", "raise", "keep_first"] = "raise",
        **kwargs,
    ) -> Self:
        """
        Chunk the data in the spool along specified dimension.

        Parameters
        ----------
        overlap
            The amount of overlap between each segment, starting with the end of
            first patch. Negative values can be used to create gaps.
        keep_partial
            If True, keep the segments which are smaller than chunk size.
            This often occurs because of data gaps or at end of chunks.
        snap_coords
            If True, snap the coords on joined patches such that the spacing
            remains constant.
        tolerance
            The number of samples a block of data can be spaced and still be
            considered contiguous.
        conflict
            {conflict_desc}
        kwargs
            kwargs are used to specify the dimension along which to chunk, eg:
            `time=10` chunks along the time axis in 10 second increments.

        Examples
        --------
        >>> import dascore as dc
        >>> from dascore.units import s
        >>>
        >>> spool = dc.get_example_spool("random_das")
        >>> # get spools with time duration of 10 seconds
        >>> time_chunked = spool.chunk(time=10, overlap=1)
        >>> # merge along time axis
        >>> time_merged = spool.chunk(time=...)

        Notes
        -----
        [`Spool.concatenate`](`dascore.BaseSpool.concatenate`) performs a
        similar operation but disregards the coordinate values.
        """

    @abc.abstractmethod
    def select(self, **kwargs) -> Self:
        """
        Sub-select parts of the spool.

        Can be used to specify dimension ranges, or unix-style matches
        on string attributes.

        Parameters
        ----------
        **kwargs
            Specifies query. Can be of the form {dim_name=(start, stop)}
            or {attr_name=query}.

        Examples
        --------
        >>> import dascore as dc
        >>> spool = dc.get_example_spool("diverse_das")
        >>> # subselect data in a particular time range
        >>> time = ('2020-01-03', '2020-01-03T00:00:10')
        >>> time_spool = spool.select(time=time)
        >>> # subselect based on matching tag parameter
        >>> tag_spool = spool.select(tag='some*')
        """

    @abc.abstractmethod
    def get_contents(self) -> pd.DataFrame:
        """
        Get a dataframe of the spool contents.

        Examples
        --------
        >>> import dascore as dc
        >>> spool = dc.get_example_spool("random_das")
        >>> df = spool.get_contents()
        """

    # Bind get_patch names as a spool method.
    get_patch_names = get_patch_names

    # --- optional methods

    def sort(self, attribute) -> Self:
        """
        Sort the Spool based on a specific attribute.

        Parameters
        ----------
        attribute
            The attribute or coordinate used for sorting. If a coordinate name
            is used, the sorting will be based on the minimum value.

        Examples
        --------
        >>> import dascore as dc
        >>> spool = dc.get_example_spool()
        >>> # sort spool based on values in time coordinate.
        >>> spool_time_sorted = spool.sort("time")
        >>> # sort spool based on values in tag
        >>> spool_tag_sorted = spool.sort("tag")
        """
        msg = f"spool of type {self.__class__} has no sort implementation"
        raise NotImplementedError(msg)

    def split(
        self,
        size: int | None = None,
        count: int | None = None,
    ) -> Generator[Self, None, None]:
        """
        Yield sub-patches based on specified parameters.

        Parameters
        ----------
        size
            The number of patches desired in each output spool. The last
            spool may have fewer patches.
        count
            The number of spools to include. If count is greater than
            the length of the spool then the output will be smaller than
            count, with one patch per spool.

        Examples
        --------
        >>> import dascore as dc
        >>> spool = dc.get_example_spool("diverse_das")
        >>> # split spool into list of spools each with 3 patches.
        >>> split = spool.split(size=3)
        >>> # split spool into 3 evenly sized (if possible) spools
        >>> split = spool.split(count=3)
        """
        msg = f"spool of type {self.__class__} has no split implementation"
        raise NotImplementedError(msg)

    def update(self, progress: PROGRESS_LEVELS = "standard") -> Self:
        """
        Updates the contents of the spool, return the updated spool.

        Parameters
        ----------
        progress
            Controls the progress bar. "standard" produces the standard
            progress bar. "basic" is a simplified version with lower refresh
            rates, best for high-latency environments, and None disables
            the progress bar.
        """
        return self

    @compose_docstring(desc=concatenate_patches.__doc__)
    def concatenate(self, check_behavior: WARN_LEVELS = "warn", **kwargs):
        """{desc}"""
        msg = f"spool of type {self.__class__} has no concatenate implementation"
        raise NotImplementedError(msg)

    def map(
        self,
        func: Callable[[dc.Patch, ...], T],
        *,
        client: ExecutorType | None = None,
        size: int | None = None,
        progress: bool = True,
        **kwargs,
    ) -> list[T]:
        """
        Map a function of all the contents of the spool.

        Parameters
        ----------
        func
            A callable which takes a patch as its first argument.
        client
            A client, or executor, which has a `map` method.
        size
            The number of patches in each spool mapped to a client.
            If not set, defaults to the number of processors on the host.
            Does nothing unless client is defined.
        progress
            If True, display a progress bar.
        **kwargs
            kwargs passed to func.

        Notes
        -----
        When a client is specified, the spool is split then passed to the
        client's map method. This is to avoid serializing loaded patches.
        See [`Spool.split`](`dascore.core.spool.BaseSpool.split`) for more
        details about the `spool_count` and `spool_size` parameters.

        Examples
        --------
        import numpy as np
        import dascore as dc

        spool = dc.get_example_spool("random_das")

        # Calculate the std for each channel in 5 second chunks
        results = (
             spool.chunk(time=5)
             .map(lambda x: np.std(x.data, axis=0))
        )
        # stack back into array. dims are (distance, time chunk)
        out = np.stack(results, axis=-1)
        """
        return _spool_map(
            self,
            func,
            client=client,
            size=size,
            progress=progress,
            **kwargs,
        )

    # Add method for stacking (adding the data arrays) patches in spool.
    stack = stack_patches

    @property
    def viz(self):
        """Raise AttributeError when Spool.viz is accessed."""
        msg = (
            "'Spool' has no 'viz' namespace. "
            "Apply 'viz' on a Patch object. "
            "(you can merge a subset of the spool into a single patch using "
            "the Chunk function. i.e., spool.chunk(time=None)[0].viz.waterfall())"
        )
        raise AttributeError(msg)


class DataFrameSpool(BaseSpool):
    """An abstract class for spools whose contents are managed by a dataframe."""

    # A dataframe which represents contents as they will be output
    _df: pd.DataFrame = CacheDescriptor("_cache", "_get_df")
    # A dataframe which shows patches in the source
    _source_df: pd.DataFrame = CacheDescriptor("_cache", "_get_source_df")
    # A dataframe of instructions for going from source_df to df
    _instruction_df: pd.DataFrame = CacheDescriptor("_cache", "_get_instruction_df")
    # kwargs for filtering contents
    _select_kwargs: Mapping | None = FrozenDict()
    # kwargs for merging patches
    _merge_kwargs: Mapping | None = FrozenDict()
    # attributes which effect merge groups for internal patches
    _group_columns = ("network", "station", "dims", "data_type", "tag")
    _drop_columns = ("patch",)

    def _get_df(self):
        """Function to get the current df."""

    def _get_source_df(self):
        """Function to get the current df."""

    def _get_instruction_df(self):
        """Function to get the current df."""

    def __init__(
        self, select_kwargs: dict | None = None, merge_kwargs: dict | None = None
    ):
        self._cache = {}
        self._select_kwargs = {} if select_kwargs is None else select_kwargs
        self._merge_kwargs = {} if merge_kwargs is None else merge_kwargs

    def _select_from_array(self, array) -> Self:
        """Create new spool with contents changed from array input."""
        if np.issubdtype(array.dtype, np.bool_):  # boolean select
            df = self._df[array]
        elif np.issubdtype(array.dtype, np.integer):
            df = self._df.iloc[array]
        else:
            msg = "Only bool or int dtypes are supported for spool array selection."
            raise ValueError(msg)
        source = self._source_df
        inst = self._instruction_df
        select_kwargs, merge_kwargs = self._select_kwargs, self._merge_kwargs
        new = self.new_from_df(
            df,
            source_df=source,
            instruction_df=inst,
            select_kwargs=select_kwargs,
            merge_kwargs=merge_kwargs,
        )
        return new

    def __getitem__(self, item) -> PatchType | BaseSpool:
        if isinstance(item, slice):  # a slice was used, return a sub-spool
            new_df = self._df.iloc[item]
            inst, source = self._instruction_df, self._source_df
            new_inst = inst[inst["current_index"].isin(new_df.index)]
            new_source = source.loc[new_inst.index]
            out = self.new_from_df(
                df=new_df,
                instruction_df=new_inst,
                source_df=new_source,
            )
        elif is_array(item):  # An array was passed use np type selection.
            return self._select_from_array(np.asarray(item))
        else:  # a single index was used, should return a single patch
            out = self._unbox_patch(self._get_patches_from_index(item))
        return out

    def __len__(self):
        return len(self._df)

    def __iter__(self):
        for ind in range(len(self._df)):
            yield self._unbox_patch(self._get_patches_from_index(ind))

    def _unbox_patch(self, patch_list):
        """Unbox a single patch from a patch list, check len."""
        assert len(patch_list) == 1
        return patch_list[0]

    def _get_patches_from_index(self, df_ind):
        """Given an index (from current df), return the corresponding patch."""
        source = self._source_df
        instruction = self._instruction_df
        # handle negative index.
        df_ind = df_ind if df_ind >= 0 else len(self._df) + df_ind
        try:
            inds = self._df.index[df_ind]
        except IndexError:
            msg = f"index of [{df_ind}] is out of bounds for spool."
            raise IndexError(msg)
        df1 = instruction[instruction["current_index"] == inds]
        assert not df1.empty
        joined = df1.join(source.drop(columns=df1.columns, errors="ignore"))
        return self._patch_from_instruction_df(joined)

    def _patch_from_instruction_df(self, joined):
        """Get the patches joined columns of instruction df."""
        out = []
        df_dict_list = self._df_to_dict_list(joined)
        expected_len = len(joined["current_index"].unique())
        for patch_kwargs in df_dict_list:
            # convert kwargs to format understood by parser/patch.select
            kwargs = _convert_min_max_in_kwargs(patch_kwargs, joined)
            patch = self._load_patch(kwargs)
            # If the limits of the source patch were not modified, we can just
            # use the select kwargs. This is important for missing coordinates
            # (NaN values) to not get trimmed out.
            if kwargs.get("_modified"):
                select_kwargs = {
                    i: v
                    for i, v in kwargs.items()
                    if i in patch.dims or i in patch.coords.coord_map
                }
            else:
                select_kwargs = self._select_kwargs
            patch: dc.Patch = patch.select(**select_kwargs)
            # its unfortunate, but currently we need to regenerate the patch
            # dict because the index doesn't carry all the dimensional info
            info = patch.attrs.flat_dump(exclude=["history"])
            info["patch"] = patch
            out.append(info)
        if len(out) > expected_len:
            out = _force_patch_merge(out, merge_kwargs=self._merge_kwargs)
        return [x["patch"] for x in out]

    def _get_dummy_dataframes(self, current):
        """
        Return dummy current, source, and instruction dataframes.

        Dummy because the source and current df are the same, so the
        instruction df is a straight mapping between the two.
        """
        source = current.copy(deep=False)  # shallow to not copy patches
        dims = get_dim_names_from_columns(source)
        cols2keep = get_column_names_from_dim(dims)
        instruction = (
            current.copy(deep=False)[cols2keep]
            .assign(
                source_index=source.index,
                current_index=source.index,
                _modified=lambda x: _column_or_value(x, "_modified", False),
            )
            .set_index("source_index")
            .sort_values("current_index")
        )
        return current, source, instruction

    def _df_to_dict_list(self, df):
        """
        Convert the dataframe to a list of dicts for iteration.

        This is significantly faster than iterating rows.
        """
        return df.to_dict("records")

    @abc.abstractmethod
    def _load_patch(self, kwargs) -> dc.Patch:
        """Given a row from the managed dataframe, return a patch."""

    @compose_docstring(doc=BaseSpool.chunk.__doc__)
    def chunk(
        self,
        overlap: numeric_types | timeable_types | None = None,
        keep_partial: bool = False,
        snap_coords: bool = True,
        tolerance: float = 1.5,
        conflict: Literal["drop", "raise", "keep_first"] = "raise",
        **kwargs,
    ) -> Self:
        """{doc}"""
        df = self._df.drop(columns=list(self._drop_columns), errors="ignore")
        chunker = ChunkManager(
            overlap=overlap,
            keep_partial=keep_partial,
            snap_coords=snap_coords,
            group_columns=self._group_columns,
            tolerance=tolerance,
            conflict=conflict,
            **kwargs,
        )
        in_df, out_df = chunker.chunk(df)
        if df.empty:
            instructions = None
        else:
            instructions = chunker.get_instruction_df(in_df, out_df)
        return self.new_from_df(
            out_df,
            source_df=self._df,
            instruction_df=instructions,
            merge_kwargs={"conflicts": conflict},
        )

    def new_from_df(
        self,
        df,
        source_df=None,
        instruction_df=None,
        select_kwargs=None,
        merge_kwargs=None,
    ):
        """Create a new instance from dataframes."""
        new = self.__class__(self)
        df_, source_, inst_ = self._get_dummy_dataframes(df)
        new._df = df
        new._source_df = source_df if source_df is not None else source_
        new._instruction_df = instruction_df if instruction_df is not None else inst_
        new._select_kwargs = dict(self._select_kwargs)
        new._select_kwargs.update(select_kwargs or {})
        new._merge_kwargs = dict(self._merge_kwargs)
        new._merge_kwargs.update(merge_kwargs or {})
        return new

    @compose_docstring(doc=BaseSpool.select.__doc__)
    def select(self, **kwargs) -> Self:
        """{doc}."""
        _, _, extra_kwargs = split_df_query(kwargs, self._df, ignore_bad_kwargs=True)
        filtered_df = adjust_segments(self._df, ignore_bad_kwargs=True, **kwargs)
        inst = adjust_segments(
            self._instruction_df,
            ignore_bad_kwargs=True,
            **kwargs,
        ).loc[lambda x: x["current_index"].isin(filtered_df.index)]
        # Determine if the instructions are the same as the source dataframe.
        out = self.new_from_df(
            filtered_df,
            source_df=self._source_df,
            instruction_df=inst,
            select_kwargs=extra_kwargs,
        )
        return out

    @compose_docstring(doc=BaseSpool.sort.__doc__)
    def sort(self, attribute) -> Self:
        """{doc}."""
        df = self._df
        inst_df = self._instruction_df

        # make sure a suitable attribute is entered
        attrs = set(df.columns)
        if attribute not in attrs:
            # make sure we can also cover coordinate names instead of the attribute
            if f"{attribute}_min" in attrs:
                attribute = f"{attribute}_min"
            else:
                msg = (
                    "Invalid attribute. " "Please use a valid attribute such as: 'time'"
                )
                raise IndexError(msg)

        # get a mapping from the old current index to the sorted ones
        sorted_df = df.sort_values(attribute).reset_index(drop=True)
        old_indices = df.index
        new_indices = np.arange(len(df))
        mapper = pd.Series(new_indices, index=old_indices)
        # swap out all the old values with new ones
        new_current_index = inst_df["current_index"].map(mapper)
        new_instruction_df = inst_df.assign(current_index=new_current_index)
        # create new spool from new dataframes
        return self.new_from_df(df=sorted_df, instruction_df=new_instruction_df)

    @compose_docstring(doc=BaseSpool.split.__doc__)
    def split(
        self,
        size: int | None = None,
        count: int | None = None,
    ) -> Generator[Self, None, None]:
        """{doc}."""
        if not ((count is not None) ^ (size is not None)):
            msg = "Spool.split requires either spool_count or spool_size."
            raise ParameterError(msg)
        start = 0
        step = int(np.ceil(len(self) / count if count else size))
        while start < len(self):
            yield self[start : start + step]
            start += step

    @compose_docstring(doc=BaseSpool.get_contents.__doc__)
    def get_contents(self) -> pd.DataFrame:
        """{doc}."""
        return self._df[filter_df(self._df, **self._select_kwargs)]

    get_patch_names = get_patch_names


class MemorySpool(DataFrameSpool):
    """A Spool for storing patches in memory."""

    # tuple of attributes to remove from table

    def __init__(self, data: PatchType | Sequence[PatchType] | None = None):
        super().__init__()
        if data is not None:
            dfs = self._get_dummy_dataframes(patches_to_df(data))
            self._df, self._source_df, self._instruction_df = dfs

    def __rich__(self):
        base = super().__rich__()
        df = self._df
        if len(df):
            t1 = df["time_min"].min() if "time_min" in df.columns else ""
            t2 = df["time_min"].max() if "time_min" in df.columns else ""
            tmin = get_nice_text(t1)
            tmax = get_nice_text(t2)
            if t1 != "" and t2 != "":
                duration = get_nice_text(t2 - t1)
            else:
                duration = ""
            base += Text(f"\n    Time Span: <{duration}> {tmin} to {tmax}")
        return base

    def _load_patch(self, kwargs) -> Self:
        """Load the patch into memory."""
        return kwargs["patch"]

    # Add specific implementation of concatenate patches.
    concatenate = _spool_up(concatenate_patches)


@singledispatch
def spool(obj: Path | str | BaseSpool | Sequence[PatchType], **kwargs) -> BaseSpool:
    """
    Create a spool from a data source.

    This is the main function for loading in DASCore.

    Parameters
    ----------
    obj
        An object from which a spool can be derived.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.downloader import fetch
    >>>
    >>> # Get a spool from a single file
    >>> single_file_path = fetch("example_dasdae_event_1.h5")
    >>> file_spool = dc.spool(single_file_path)
    >>>
    >>> # get a spool from a directory of files
    >>> directory_path = fetch("example_dasdae_event_1.h5").parent
    >>> directory_spool = dc.spool(directory_path)
    >>>
    >>> # get a spool from a single patch
    >>> patch = dc.get_example_patch()
    >>> spool = dc.spool(patch)
    """
    msg = f"Could not get spool from: {obj}"
    raise ValueError(msg)


@spool.register(str)
@spool.register(Path)
def _spool_from_str(path, **kwargs):
    """Get a spool from a path."""
    path = Path(path)
    # A directory was passed, create Directory Spool
    if path.is_dir():
        from dascore.clients.dirspool import DirectorySpool

        return DirectorySpool(path, **kwargs)
    # A single file was passed. If the file format supports quick scanning
    # Return a FileSpool (lazy file reader), else return DirectorySpool.
    elif path.exists():  # a single file path was passed.
        _format, _version = dc.get_format(path, **kwargs)
        formatter = dc.io.FiberIO.manager.get_fiberio(format=_format, version=_version)
        if formatter.implements_scan:
            from dascore.clients.filespool import FileSpool

            return FileSpool(path, _format, _version)

        else:
            return MemorySpool(dc.read(path, _format, _version))
    else:
        msg = (
            f"could not get spool from argument: {path}. "
            f"If it is a path, it may not exist."
        )
        raise InvalidSpoolError(msg)


@spool.register(BaseSpool)
def _spool_from_spool(spool, **kwargs):
    """Return a spool from a spool."""
    return spool


@spool.register(list)
@spool.register(tuple)
def _spool_from_patch_list(patch_list, **kwargs):
    """Return a spool from a sequence of patches."""
    return MemorySpool(patch_list)


@spool.register(dc.Patch)
def _spool_from_patch(patch):
    """Get a spool from a single patch."""
    return MemorySpool([patch])



"""Processing operations that have much to do with coordinates."""

from __future__ import annotations

from collections.abc import Collection

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from dascore.constants import PatchType, select_values_description
from dascore.core.coords import BaseCoord
from dascore.exceptions import CoordError, ParameterError, PatchError
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import get_parent_code_name
from dascore.utils.patch import patch_function


@patch_function()
def snap_coords(patch: PatchType, *coords, reverse=False):
    """
    Snap coordinates to evenly spaced samples.

    This ensures all of the specified coordinates are evenly spaced and
    monotonic. First, the patch is sorted along specified coordinates,
    then coordinates are assumed evenly-sampled from their min to max value.
    Doing this can introduce some error since the coordinate labels are moved
    and the data (apart from the sorting) are left unchanged.
    Consider using [interpolate](`dascore.Patch.interpolate`) for a more expensive
    but more accurate linear interpolation.

    Parameters
    ----------
    *coords
        Used to specify the dimension names to convert to CoordRanges. If not
        specified convert all dimensional coordinates.
    reverse
        If True, reverse the sorting of the coordinates.

    Examples
    --------
    >>> import dascore as dc
    >>> # get an example patch which has unevenly sampled coords time, distance
    >>> patch = dc.get_example_patch("wacky_dim_coords_patch")
    >>> # snap time dimension
    >>> time_snap = patch.snap_coords("time")
    >>> # snap the distance dimension
    >>> dist_snap = patch.snap_coords("distance")
    """
    cman, data = patch.coords.snap(*coords, array=patch.data, reverse=reverse)
    return patch.new(data=data, coords=cman)


@patch_function()
def sort_coords(patch: PatchType, *coords, reverse=False):
    """
    Sort one or more coordinates.

    Sorts the specified coordinates in the patch. An error will be raised
    if the coordinates have overlapping dimensions since it may not be
    possible to sort each. An error is also raised in any of the coordinates
    are multidimensional.

    Parameters
    ----------
    *coords
        Used to specify the coordinates to sort.
    reverse
        If True, sort in descending order, else ascending.

    Examples
    --------
    >>> import dascore as dc
    >>> # get an example patch which has unevenly sampled coords time, distance
    >>> patch = dc.get_example_patch("wacky_dim_coords_patch")
    >>> # sort time coordinate (dimension) in ascending order
    >>> time_snap = patch.sort_coords("time")
    >>> assert time_snap.coords.coord_map['time'].sorted
    >>> # sort distance coordinate (dimension) in descending order
    >>> dist_snap = patch.sort_coords("distance", reverse=True)
    >>> assert dist_snap.coords.coord_map['distance'].reverse_sorted
    """
    cman, data = patch.coords.sort(*coords, array=patch.data, reverse=reverse)
    return patch.new(data=data, coords=cman)


def get_coord(
    self: PatchType,
    name: str,
    require_sorted: bool = False,
    require_evenly_sampled: bool = False,
) -> BaseCoord:
    """
    Get a managed coordinate from the patch.

    Parameters
    ----------
    name
        The name of the coordinate to fetch from the patch.
    require_sorted
        If True, require the coordinate to be sorted or raise Error.
    require_evenly_sampled
        If True, require the coordinate to be evenly sampled or raise Error.

    Raises
    ------
    [`CoordError`](`dascore.exceptions.CoordError`) if the coordinate does
    not exist or does not meet the imposed requirements.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Get the the distance coordinate from the patch.
    >>> distance = patch.get_coord("distance")
    >>>
    >>> # Get the time coordinate from the patch, raise CoordError if it
    >>> # is not evenly sampled.
    >>> time = patch.get_coord("time", require_evenly_sampled=True)

    See Also
    --------
    [get_array](`dascore.Patch.get_array`).

    """
    if (coord := self.coords.coord_map.get(name)) is None:
        coords = sorted(self.coords.coord_map)
        msg = f"Coordinate '{name}' not found in Patch coordinates: {coords}"
        raise CoordError(msg)
    if require_evenly_sampled and coord.step is None:
        extra = f"as required by {get_parent_code_name()}"  # adds caller name
        msg = f"Coordinate {name} is not evenly sampled {extra}"
        raise CoordError(msg)
    if require_sorted and not (coord.sorted or coord.reverse_sorted):
        extra = f"as required by {get_parent_code_name()}"  # adds caller name
        msg = f"Coordinate {name} is not sorted {extra}"
        raise CoordError(msg)
    return coord


def get_array(
    self: PatchType,
    name: str | None = None,
    require_sorted: bool = False,
    require_evenly_sampled: bool = False,
) -> BaseCoord:
    """
    Get an array associated with patch data or a coordinate.

    Parameters
    ----------
    name
        The name of the coordinate to fetch. If None return patch data.
    require_sorted
        If True, require the coordinate to be sorted or raise Error.
    require_evenly_sampled
        If True, require the coordinate to be evenly sampled or raise Error.

    Raises
    ------
    [`CoordError`](`dascore.exceptions.CoordError`) if the coordinate does
    not exist or does not meet the imposed requirements.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Get the patch data array.
    >>> data = patch.get_array()  # same as patch.data
    >>>
    >>> # Get an array of distance values
    >>> distance_array = patch.get_array("distance")
    >>>
    >>> # Get an array of time values. Raise an error if they arent sorted.
    >>> time_array = patch.get_array("time", require_sorted=True)

    See Also
    --------
    [Patch.get_coord](`dascore.Patch.get_coord`)
    """
    if name is None:
        return self.data
    coord = get_coord(
        self,
        name,
        require_sorted=require_sorted,
        require_evenly_sampled=require_evenly_sampled,
    )
    return coord.data


@patch_function()
def rename_coords(self: PatchType, **kwargs) -> PatchType:
    """
    Rename coordinate of Patch.

    Parameters
    ----------
    **kwargs
        The mapping from old names to new names

    Examples
    --------
    >>> import dascore as dc
    >>> pa = dc.get_example_patch()
    >>> # rename dim "distance" to "fragrance"
    >>> pa2 = pa.rename_coords(distance='fragrance')
    >>> assert 'fragrance' in pa2.dims
    """
    new_coord = self.coords.rename_coord(**kwargs)
    attrs = self.attrs.rename_dimension(**kwargs)
    return self.new(coords=new_coord, dims=new_coord.dims, attrs=attrs)


@patch_function()
def update_coords(self: PatchType, **kwargs) -> PatchType:
    """
    Update the coordiantes of a patch.

    Will either add new coordinates, or update existing ones.

    Parameters
    ----------
    kwargs
        The name of the coordinate (key) and coordinate values. Values
        can either be a sequence (eg array) or a single int. If an int
        is used it will create a non-coord.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> pa = dc.get_example_patch()
    >>>
    >>> # Add 1 to all distance coords
    >>> new_dist = pa.coords.get_array('distance') + 1
    >>> pa2 = pa.update_coords(distance=new_dist)
    >>> assert np.allclose(pa2.coords.get_array('distance'), new_dist)
    """
    new_coord = self.coords.update(**kwargs)
    return self.new(coords=new_coord, dims=new_coord.dims)


@patch_function()
def drop_coords(self: PatchType, *coords: str | Collection[str]) -> PatchType:
    """
    Update the coordiantes of a patch.

    Will either add new coordinates, or update existing ones.

    Parameters
    ----------
    *coords
        One or more coordinates to drop.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> pa = dc.get_example_patch("random_patch_with_lat_lon")
    >>> # Drop non-dimensional coordinate latitude
    >>> pa_no_lat = pa.drop_coords("latitude")
    """
    if dim_coords := set(coords) & set(self.dims):
        msg = f"Cannot drop dimensional coordinates: {dim_coords}"
        raise ParameterError(msg)
    new_coord, data = self.coords.drop_coords(*coords, array=self.data)
    return self.new(coords=new_coord, dims=new_coord.dims, data=data)


@patch_function()
def drop_private_coords(self: PatchType) -> PatchType:
    """
    Drop all private coords in the patch.

    Parameters
    ----------
    self
        Patch

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> pa = (
    ...     dc.get_example_patch("random_das")
    ...     .update_coords(_private=(None, np.array([1,2,3])))
    ... )
    >>> pa_no_private = pa.drop_private_coords()
    >>> assert "_private" not in pa_no_private.coords.coord_map
    """
    new_coord, data = self.coords.drop_private_coords(array=self.data)
    return self.new(coords=new_coord, dims=new_coord.dims, data=data)


@patch_function()
def make_broadcastable_to(
    self: PatchType,
    shape: tuple[int, ...],
    drop_coords=False,
) -> PatchType:
    """
    Update the coordiantes of a patch.

    Will either add new coordinates, or update existing ones.

    Parameters
    ----------
    shape
        The new shape the patch should be able to broadcast with.
    drop_coords
        If True, drop coords that need to be broadcasted up, otherwise
        only NonCoordinate dimensions can change shape.

    Examples
    --------
    >>> import dascore as dc
    >>> pa = dc.get_example_patch("random_das")
    >>> # Get a patch with non-coordinate dimensions
    >>> patch = pa.mean()
    >>> out = patch.make_broadcastable_to(shape=(2, 3))
    >>> assert out.shape == (2, 3)
    """
    coords, data = self.coords.make_broadcastable_to(
        shape, array=self.data, drop_coords=drop_coords
    )
    return self.new(coords=coords, data=data)


@patch_function(history="method_name")
def coords_from_df(
    self: PatchType,
    dataframe: pd.DataFrame,
    units: dict[str, str] | None = None,
    extrapolate: bool = False,
) -> PatchType:
    """
    Update non-dimensional coordinate of a patch using a dataframe.

    Parameters
    ----------
    dataframe
        Table with a column matching in title to one of patch.dims along with other
        coordinates to associate with dimension. Example one column matching distance
        axis and then latitude and longitude attached to the distances.
    units
        Dictionary mapping column name in dataframe to its units.
    extrapolate
        If True, extrapolate outside provided range in dataframe.

    Examples
    --------
    >>> import dascore as dc
    >>> import pandas as pd
    >>> # get example patch and create example dataframe
    >>> pa = dc.get_example_patch()
    >>> distance = pa.coords.get_array("distance")[::10]
    >>> df = pd.DataFrame(distance, columns=['distance'])
    >>> df['x'] = df['distance'] * 3 + 10
    >>> df['y'] = df['distance'] * 2.5 - 10
    >>> # attach dataframe to patch, interpolating when needed. This
    >>> # adds coordinates x and y which are associated with dimension distance.
    >>> patch_with_coords = pa.coords_from_df(df)

    Notes
    -----
    * Exactly one of the column names in the dataframe must map to one of
      the patch.dims. This will either add new coordinates, or update existing
      ones if they already exist.

    * This function uess linear extrapolation between the nearest two points
      to get values in patch coords that aren't in the dataframe.

    """
    # match dataframe headings to dims
    anchor_dim = set(self.dims) & set(dataframe.columns)
    if len(anchor_dim) != 1:
        msg = "Exactly one column has to match with an existing dimension"
        raise ParameterError(msg)

    # Get coordinates of axis being updated
    anchor_dim = next(iter(anchor_dim))
    coords = self.coords
    axis_coords = coords.get_array(anchor_dim)

    # make a dictionary from coordinates("(axis, coordinate array)") as input to
    # update_coords
    # coordinate array is an interpolation to match existing coords being updated
    new_coords = {}

    for coord in set(dataframe.columns) - {anchor_dim}:
        if extrapolate:
            f = interp1d(
                pd.to_numeric(dataframe[anchor_dim]),
                pd.to_numeric(dataframe[coord]),
                fill_value="extrapolate",
            )
            new_coords[coord] = (anchor_dim, f(axis_coords))
        else:
            new_coords[coord] = (
                anchor_dim,
                np.interp(
                    axis_coords,
                    pd.to_numeric(dataframe[anchor_dim]),
                    pd.to_numeric(dataframe[coord]),
                    left=float("nan"),
                    right=float("nan"),
                ),
            )

    out = self.update_coords.func(self, **new_coords)

    if units is not None:
        out = out.convert_units.func(out, **units)

    return out


@patch_function(history=None)
@compose_docstring(select_params=select_values_description)
def select(
    patch: PatchType, *, copy=False, relative=False, samples=False, **kwargs
) -> PatchType:
    """
    Return a subset of the patch.

    {select_params}

    Parameters
    ----------
    patch
        The patch object.
    copy
        If True, copy the resulting data. This is needed so the old
        array can get gc'ed and memory freed.
    relative
        If True, select ranges are relative to the start of coordinate, if
        possitive, or the end of the coordinate, if negative.
    samples
        If True, the query meaning is in samples.
    **kwargs
        Used to specify the coordinate on which data are selected.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> from dascore.examples import get_example_patch
    >>> patch = get_example_patch()
    >>> # select meters 50 to 300
    >>> new_distance = patch.select(distance=(50, 300))
    >>> # select channels less than 300
    >>> lt_dist = patch.select(distance=(..., 300))
    >>> # select time (1 second from start to -1 second from end)
    >>> t1 = patch.attrs.time_min + dc.to_timedelta64(1)
    >>> t2 = patch.attrs.time_max - dc.to_timedelta64(1)
    >>> new_time1 = patch.select(time=(t1, t2))
    >>> # this can be accomplished more simply using the relative keyword
    >>> new_time2 = patch.select(time=(1, -1), relative=True)
    >>> # filter 1 second from start time to 3 seconds from start time
    >>> new_time3 = patch.select(time=(1, 3), relative=True)
    >>> # filter 6 second from end time to 1 second from end time
    >>> new_time4 = patch.select(time=(-6, -1), relative=True)
    >>> # Select first 10 distance indices
    >>> new_distance1 = patch.select(distance=(..., 10), samples=True)
    >>> # Select last time row/column
    >>> new_distance2 = patch.select(time=-1, samples=True)
    >>> # only include certain rows/columns based on a boolean array.
    >>> time = patch.get_array("time")
    >>> new_time_5 = patch.select(time=time>time[2])
    >>> # Select only specific values along a dimension
    >>> distance = patch.get_array("distance")
    >>> new_distance_3 = patch.select(distace=distance[1::2])

    Notes
    -----
    - It is important to remember select will not change the order of the
      patch, only fiter values. If the order of the patch should change, or
      multiple rows/columns need to be repeated,
      See [`Patch.order`](`dascore.Patch.order`).

    """
    new_coords, data = patch.coords.select(
        **kwargs,
        array=patch.data,
        relative=relative,
        samples=samples,
    )
    # no slicing was performed, just return original.
    if data.shape == patch.data.shape:
        return patch
    if copy:
        data = data.copy()
    return patch.new(data=data, coords=new_coords)


@patch_function(history=None)
@compose_docstring(select_params=select_values_description)
def order(
    patch: PatchType, *, copy=False, relative=False, samples=False, **kwargs
) -> PatchType:
    """
    Re-order the patch contents based on coordinate values or indices.

    Parameters
    ----------
    patch
        The patch object.
    copy
        If True, copy the resulting data. This is needed so the old
        array can get gc'ed and memory freed.
    relative
        If True, order values are relative to the start/end of the coordinates.
    samples
        If True, the
    **kwargs
        Used to specify the coordinate and values on which the coordinates
        are ordered.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.examples import get_example_patch
    >>> patch = get_example_patch()
    >>> # Sub-select only a section of the distance and ensure order.
    >>> dist = patch.get_array("distance")
    >>> new_dist = dist[1:5][::-1]
    >>> patch_1 = patch.order(distance=new_dist)
    >>> # Get duplicate the first time row or column
    >>> patch_2 = patch.order(time=[0, 0, 0], samples=True)

    Notes
    -----
    - This function is similar to [`Patch.select`](`dascore.Patch.select`)
      but it will also change the patch order to match the inputs exactly.
      If there are repeated values in the requsted values or in the patch
      coordinate arrays, the data will end up being repeated as well.
    """
    new_coords, data = patch.coords.order(
        **kwargs,
        array=patch.data,
        relative=relative,
        samples=samples,
    )
    if copy:
        data = data.copy()
    return patch.new(data=data, coords=new_coords)


@patch_function(history=None)
def transpose(self: PatchType, *dims: str) -> PatchType:
    """
    Transpose the data array to any dimension order desired.

    Parameters
    ----------
    *dims
        Dimension names which define the new data axis order.

    Examples
    --------
    >>> import dascore # import dascore library
    >>> pa = dascore.get_example_patch() # generate example patch
    >>> # transpose the time and data array dimensions in the example patch
    >>> out = dascore.proc.transpose(pa,"time", "distance")
    """
    dims = tuple(dims)
    old_dims = self.coords.dims
    new_coord = self.coords.transpose(*dims)
    new_dims = new_coord.dims
    axes = tuple(old_dims.index(x) for x in new_dims)
    new_data = np.transpose(self.data, axes)
    return self.new(data=new_data, coords=new_coord)


@patch_function(history=None)
def append_dims(patch: PatchType, *empty_dims, **dim_kwargs) -> PatchType:
    """
    Insert dimensions at the end of the patch.

    Parameters
    ----------
    empty_dims
        Used to pass the name of empty dimensions.
    dim_kwargs
        Used to pass keys (new dim names) and values. Values can either be
        an int specifying the length of the new dimension or a sequence
        specifying the coordinate values. If an int is used, the new dimension
        will be a non-coordinate dimension.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()

    >>> # Add two dummy dimensions to patch named "end" and "stop"
    >>> new = patch.append_dims("end", "stop")

    >>> # Add a dummy dimension called "face" to end of patch
    >>> # which has a coordinate value of [1].
    >>> new = patch.append_dims(face=[1])

    >>> # Same thing as above, but with a larger coords which broadcasts
    >>> # the data to shape appropriate to mach coordinates.
    >>> new = patch.append_dims(face=[1, 2])

    >>> # Add a dummy dimension of length 3 to end of patch.
    >>> # the data to shape appropriate to mach coordinates.
    >>> new = patch.append_dims(face=3)

    Notes
    -----
    - This tries to be more simple than numpy and xarray's expand_dims.
    - Use [`Patch.transpose`](`dascore.Patch.transpose`) to re-arrange dimensions.
    - If dimension with the same name already exists nothing will happen.
    """
    dim_dict = {x: 1 for x in empty_dims}
    dim_dict.update(dim_kwargs)
    # Remove duplicate dims and convert non ints to arrays.
    kwargs = {
        i: (i, np.atleast_1d(v) if not isinstance(v, int) else v)
        for i, v in dim_dict.items()
        if i not in patch.dims
    }
    # Nothing to do.
    if not kwargs:
        return patch
    ndim = patch.ndim
    # First get data with empty dimensions
    insert_inds = [x + ndim for x in range(len(kwargs))]
    data = np.expand_dims(patch.data, insert_inds)
    shapes = list(data.shape)
    for ind, (_, cdata) in zip(insert_inds, kwargs.values()):
        shapes[ind] = cdata if isinstance(cdata, int) else len(cdata)
    data = np.broadcast_to(data, shapes)
    coords = patch.coords.update(**kwargs)
    return patch.update(data=data, coords=coords)


@patch_function()
def squeeze(self: PatchType, dim=None) -> PatchType:
    """
    Return a new object with len one dimensions flattened.

    Parameters
    ----------
    dim
        Selects a subset of the length one dimensions. If a dimension
        is selected with length greater than one, an error is raised.
        If None, all length one dimensions are squeezed.
    """
    coords = self.coords.squeeze(dim)
    axis = None if dim is None else self.coords.dims.index(dim)
    data = np.squeeze(self.data, axis=axis)
    return self.new(data=data, coords=coords)


@patch_function()
def add_distance_to(
    patch: PatchType, origin: pd.Series, ord=None, prefix: str = "origin"
) -> PatchType:
    """
    Calculate the distance to "origin" and create new coordinate.

    A new coordinate called `origin_distance` (or another name controlled
    by the pre-fix argument) is added to the output patch to specify the
    exact distance. Coordinates representing the origin location
    (eg origin_x, origin_y, origin_z) are also added as non-associated
    coordinates.

    Parameters
    ----------
    patch
        The patch object which contains some overlap in coordiantes as
        index names in origin.
    origin
        A series which contains index names that overlap with patch coordinates.
        All the referenced coordinates must be associated with the same
        dimension.
    ord
        Controls the norm type. Default is Frobenius norm, see the norm
        function of numpy.linalg for supported options.
    prefix
        The prefix name for the added coordinates and attributes.

    Examples
    --------
    >>> import pandas as pd
    >>>
    >>> import dascore as dc
    >>>
    >>> # Add a coordinate specifying the distance to a theoretical shot.
    >>> shot = pd.Series({"x": 10, "y": 10, "z": 0})
    >>> patch = dc.get_example_patch("random_patch_with_xyz")
    >>> patch_with_origin_dist = patch.add_distance_to(shot)
    >>> # Now the new coordinates of distance and shot origin exist.
    >>> dist = patch_with_origin_dist.get_array("origin_distance")
    >>> origin_x = patch_with_origin_dist.get_array("origin_x")
    >>>
    >>> # Of course, the new coordinate can be used for sorting.
    >>> sorted_patch = patch_with_origin_dist.sort_coords("origin_distance")
    """
    # Ensure all index values are represented in coord map.
    if missing_coords := (set(origin.index) - set(patch.coords.coord_map)):
        msg = f"Indices {missing_coords} are not patch coordinates."
        raise PatchError(msg)
    # Ensure all coordinates have the same associated dimension.
    associated_dims = {patch.coords.dim_map[x] for x in origin.index}
    if len(associated_dims) > 1:
        dims = {i: v for i, v in patch.coords.dim_map.items() if i in origin.index}
        msg = (
            "All coordinate must be associated with the same dimension to "
            f"calculate distance. Relevant dimension mappings are {dims}"
        )
        raise PatchError(msg)
    # Create 2d arrays from coords and origin.
    coord_array = np.stack([patch.get_array(x) for x in origin.index], axis=1)
    origin_array = np.atleast_2d(origin.values)
    # Translate coords to origin and take norm.
    distance = np.linalg.norm(origin_array - coord_array, axis=1, ord=ord)
    # Add attrs and coords to new patch
    dims = next(iter(associated_dims))
    new_coords = {f"{prefix}_{i}": (None, np.atleast_1d(v)) for i, v in origin.items()}
    new_coords[f"{prefix}_distance"] = (dims, distance)
    out = patch.update_coords.func(patch, **new_coords)
    return out



"""
Module containing patch processing routines.
"""
from __future__ import annotations

import dascore.proc.aggregate as agg
from .basic import *  # noqa
from .coords import *  # noqa
from .correlate import correlate, correlate_shift
from .detrend import detrend
from .filter import median_filter, pass_filter, sobel_filter, savgol_filter, gaussian_filter, slope_filter, notch_filter
from .resample import decimate, interpolate, resample
from .rolling import rolling
from .taper import taper, taper_range
from .units import convert_units, set_units, simplify_units
from .whiten import whiten



"""
Tests for xml utilities.
"""

from dascore.utils.xml import xml_to_dict


class TestXMLtoDict:
    """Tests for converting XML to dictionaries."""

    def test_single_element(self):
        """Test conversion of XML with a single element."""
        xml_string = "<root>Hello</root>"
        assert xml_to_dict(xml_string) == "Hello"

    def test_multiple_elements(self):
        """Test conversion of XML with multiple elements."""
        xml_string = "<root><a>1</a><b>2</b></root>"
        expected_dict = {"a": "1", "b": "2"}
        assert xml_to_dict(xml_string) == expected_dict

    def test_nested_elements(self):
        """Test conversion of XML with nested elements."""
        xml_string = "<root><a><b>1</b></a></root>"
        expected_dict = {"a": {"b": "1"}}
        assert xml_to_dict(xml_string) == expected_dict

    def test_multiple_nested_elements(self):
        """Test conversion of XML with multiple nested elements."""
        xml_string = "<root><a><b>1</b></a><c><d>2</d></c></root>"
        expected_dict = {"a": {"b": "1"}, "c": {"d": "2"}}
        assert xml_to_dict(xml_string) == expected_dict

    def test_elements_repeated_twice(self):
        """Test conversion of XML with repeated elements."""
        xml_string = "<root><a>1</a><a>2</a></root>"
        expected_dict = {"a": ["1", "2"]}
        assert xml_to_dict(xml_string) == expected_dict

    def test_repeated_elements(self):
        """Test conversion of XML with repeated elements."""
        xml_string = "<root><a>1</a><a>2</a><a>3</a></root>"
        expected_dict = {"a": ["1", "2", "3"]}
        assert xml_to_dict(xml_string) == expected_dict



"""
Core modules for AP sensing support.
"""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.constants import opt_timeable_types
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader

from .utils import _get_attrs_dict, _get_patch, _get_version_string


class APSensingPatchAttrs(dc.PatchAttrs):
    """Patch Attributes for AP sensing."""

    gauge_length: float = np.nan
    radians_to_nano_strain: float = np.nan


class APSensingV10(FiberIO):
    """Support for APSensing V 10."""

    name = "APSensing"
    preferred_extensions = ("hdf5", "h5")
    version = "10"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """
        Return format name and version string if AP sensing, else False.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        version_str = _get_version_string(resource)
        if version_str:
            return self.name, version_str

    def scan(self, resource: H5Reader, **kwargs) -> list[dc.PatchAttrs]:
        """Scan an AP sensing file, return summary info about the contents."""
        file_version = _get_version_string(resource)
        extras = {
            "path": resource.filename,
            "file_format": self.name,
            "file_version": str(file_version),
        }
        attrs = _get_attrs_dict(resource)
        attrs.update(extras)
        return [APSensingPatchAttrs(**attrs)]

    def read(
        self,
        resource: H5Reader,
        time: tuple[opt_timeable_types, opt_timeable_types] | None = None,
        distance: tuple[float | None, float | None] | None = None,
        **kwargs,
    ) -> dc.BaseSpool:
        """Read a single file with APSensing data inside."""
        patches = _get_patch(
            resource, time=time, distance=distance, attr_cls=APSensingPatchAttrs
        )
        return dc.spool(patches)



"""Utilities for Febus."""

from __future__ import annotations

from collections import namedtuple

import numpy as np

import dascore as dc
from dascore.core import get_coord, get_coord_manager
from dascore.core.coordmanager import CoordManager
from dascore.utils.misc import (
    _maybe_unpack,
    broadcast_for_index,
    maybe_get_items,
    unbyte,
)

# --- Getting format/version

_FebusSlice = namedtuple(
    "FebusSlice",
    ["group", "group_name", "source", "source_name", "zone", "zone_name", "data_name"],
)


def _flatten_febus_info(fi) -> tuple[_FebusSlice, ...]:
    """
    Given a febus file, return a tuple of named tuples with key info.

    This flattens the iteration nesting to a single layer.
    """
    out = []
    for group_name, group in fi.items():
        for source_name, source in group.items():
            for zone_name, zone in source.items():
                # Skip time dataset (we only want zone groups).
                if zone_name == "time":
                    continue
                # get dataset name (not always StrainRate for older data)
                possible_ds_names = list(zone.keys())
                assert len(possible_ds_names) == 1
                data_name = possible_ds_names[0]
                zlice = _FebusSlice(
                    group, group_name, source, source_name, zone, zone_name, data_name
                )
                out.append(zlice)
    return tuple(out)


def _get_febus_version_str(hdf_fi) -> str:
    """Return the version string for febus file."""
    # Define a few root attrs that act as a "fingerprint"
    # all Febus DAS files have folders that start with fa (I hope).
    inst_keys = sorted(hdf_fi.keys())
    expected_source_attrs = {
        "AmpliPower",
        "Hostname",
        "WholeExtent",
        "SamplingRate",
    }
    # iterate instrument keys
    is_febus = all([x.startswith("fa") for x in inst_keys])
    # Version 1, or what I think is version one (eg Valencia PubDAS data)
    # did not include a Version attr in Source dataset, so we use that as
    # the default.
    version = "1"
    for inst_key in inst_keys:
        inst = hdf_fi[inst_key]
        source_keys = set(inst.keys())
        is_febus = is_febus and all(x.startswith("Source") for x in source_keys)
        for source_key in source_keys:
            source = inst[source_key]
            # If the version is set in a Source use that version.
            # Hopefully this is the file version...
            version = unbyte(source.attrs.get("Version", version)).split(".")[0]
            is_febus = is_febus and expected_source_attrs.issubset(set(source.attrs))
    if inst_keys and is_febus:
        return version
    return ""


def _get_febus_attrs(feb: _FebusSlice) -> dict:
    """Get non-coordinate attrs from febus slice."""
    zone_attrs = feb.zone.attrs
    attr_mapping = {
        "GaugeLength": "gauge_length",
        "PulseWidth": "pulse_width",
        "Version": "folog_a1_software_version",
    }
    out = maybe_get_items(zone_attrs, attr_mapping, unpack_names=set(attr_mapping))
    out["group"] = feb.group_name
    out["source"] = feb.source_name
    out["zone"] = feb.zone_name
    out["schema_version"] = out.get("folog_a1_software_version", "").split(".")[0]
    out["dims"] = ("time", "distance")
    return out


def _get_time_overlap_samples(feb, data_shape):
    """Determine the number of redundant samples in the time dimension."""
    time_step = feb.zone.attrs["Spacing"][1] / 1_000  # value in ms, convert to s.
    block_time = _maybe_unpack(1 / (feb.zone.attrs["BlockRate"] / 1_000))
    # Since the data have overlaps in each block's time dimension, we need to
    # trim the overlap off the time dimension to avoid having to merge blocks later.
    # However, sometimes the "BlockOverlap" is wrong, so we calculate it
    # manually here.
    expected_samples = int(np.round(block_time / time_step))
    excess_rows = data_shape[1] - expected_samples
    assert (
        excess_rows % 2 == 0
    ), "excess rows must be symmetric to distribute on both ends"
    return excess_rows


def _get_time_coord(feb):
    """Get the time coordinate contained in the febus slice."""
    time = feb.source["time"]
    # In older version time shape is different, always grab first element.
    first_slice = tuple(0 for _ in time.shape)
    t_0 = time[first_slice]
    # Data dimensions are block_index, time, distance
    data_shape = feb.zone[feb.data_name].shape
    n_blocks = data_shape[0]
    # Get spacing between time samples (in s) and the total time of each block.
    time_step = feb.zone.attrs["Spacing"][1] / 1_000  # value in ms, convert to s.
    excess_rows = _get_time_overlap_samples(feb, data_shape)
    total_time_rows = (data_shape[1] - excess_rows) * n_blocks
    # Get origin info, these are offsets from time to get to the first simple
    # of the block. These should always be non-positive.
    time_offset = feb.zone.attrs["Origin"][1] / 1_000  # also convert to s
    assert time_offset <= 0, "time offset must be non positive"
    # Get the start/stop indices for the zone. We assume zones never sub-slice
    # time (only distance) but assert that here.
    extent = feb.zone.attrs["Extent"]
    assert (extent[3] - extent[2] + 1) == data_shape[1], "Cant handle sub time zones"
    # Create time coord
    # Need to account for removing overlap times.
    total_start = t_0 + time_offset + (excess_rows // 2) * time_step
    total_end = total_start + total_time_rows * time_step
    time_coord = get_coord(
        start=dc.to_datetime64(total_start),
        stop=dc.to_datetime64(total_end),
        step=dc.to_timedelta64(time_step),
    )
    # Note: we have found some files in which the sampling rate is 1/3e-4
    # because we use datetime64 we lose some precision which has caused
    # slight differences in shape of the patch.
    out = time_coord.change_length(total_time_rows)
    return out


def _get_distance_coord(feb):
    """Get the distance coordinate associated with febus slice."""
    data_shape = feb.zone[feb.data_name].shape
    total_distance_inds = data_shape[2]
    # Get spacing between channels (in m)
    distance_step = feb.zone.attrs["Spacing"][0]
    # Get origin info, these are absolute for distance.
    distance_origin = feb.zone.attrs["Origin"][0]
    # Get the start/stop indicies for the zone
    extent = feb.zone.attrs["Extent"]
    dist_ids = (extent[0], extent[1])
    # Create distance coord
    # Need to account for removing overlap times.
    start = dist_ids[0] * distance_step + distance_origin
    stop = start + total_distance_inds * distance_step
    dist_coord = get_coord(
        start=start,
        stop=stop,
        step=distance_step,
        units="m",
    )
    return dist_coord.change_length(total_distance_inds)


def _get_febus_coord_manager(feb: _FebusSlice) -> CoordManager:
    """Get a coordinate manager for febus slice."""
    coords = dict(
        time=_get_time_coord(feb),
        distance=_get_distance_coord(feb),
    )
    cm = get_coord_manager(coords=coords, dims=("time", "distance"))
    return cm


def _yield_attrs_coords(fi) -> tuple[dict, CoordManager]:
    """Scan a febus file, return metadata."""
    febuses = _flatten_febus_info(fi)
    for febus in febuses:
        attr = _get_febus_attrs(febus)
        cm = _get_febus_coord_manager(febus)
        yield attr, cm, febus


def _get_data_new_cm(cm, febus, distance=None, time=None):
    """
    Get the data from febus file, maybe filtering on time/distance.

    This is a bit more complicated since the febus data are stored in a 3d array,
    but we want a 2d output.
    """

    def _get_start_end_time_array(time_coord, total_time_rows, data_shape, time):
        """Get a 2d array where columns are start/end times for each block."""
        block_count = data_shape[0]
        block_duration = total_time_rows * time_coord.step
        start = (
            np.arange(block_count) * block_duration + time_coord.step
        ) + time_coord.min()
        end = start + block_duration
        return np.stack([start, end], axis=-1)

    def _get_time_filtered_data(data, t_start_end, time, total_slice, time_coord):
        """Get new data array filtered from time query."""
        assert len(time) == 2
        t1, t2 = time
        # block for which all data are needed.
        in_time = np.ones(len(t_start_end), dtype=bool)
        if t1 is not None and t1 is not ...:
            in_time = np.logical_and(in_time, ~(t_start_end[:, 1] < t1))
        if t2 is not None and t2 is not ...:
            in_time = np.logical_and(in_time, ~(t_start_end[:, 0] > t2))
        times = t_start_end[in_time]
        tmin = times[:, 0].min()
        tmax = times[:, 1].max()
        # get start/stop indexes for complete blocks
        start = np.argmax(in_time)
        stop = np.argmax(np.cumsum(in_time))
        total_slice[0] = slice(start, stop)
        # load data from disk.
        data_2d = data[tuple(total_slice)].reshape(-1, data.shape[-1])
        # Next, get mew time coord and slice.
        new_coord, time_slice = (
            get_coord(min=tmin, max=tmax, step=time_coord.step)
            .change_length(len(data_2d))
            .select((t1, t2))
        )
        return data_2d[time_slice], new_coord

    dist_coord, time_coord = cm.coord_map["distance"], cm.coord_map["time"]
    data = febus.zone[febus.data_name]
    data_shape = data.shape
    skip_rows = _get_time_overlap_samples(febus, data_shape) // 2
    # Need to handle case where excess_rows == 0
    data_slice = slice(skip_rows, -skip_rows if skip_rows else None)
    total_slice = list(broadcast_for_index(3, 1, data_slice))
    total_time_rows = data_shape[1] - 2 * skip_rows
    if distance:
        dist_coord, total_slice[2] = dist_coord.select(distance)
    if time:  # need to sub-select blocks to get data we are after.
        t_start_end = _get_start_end_time_array(
            time_coord, total_time_rows, data_shape, time
        )
        data, time_coord = _get_time_filtered_data(
            data, t_start_end, time, total_slice, time_coord
        )
    else:  # no need to mess with blocks, all time is selected
        data_3d = data[tuple(total_slice)]
        data = data_3d.reshape(-1, data_3d.shape[2])
    cm = get_coord_manager({"time": time_coord, "distance": dist_coord}, dims=cm.dims)
    return data, cm


def _read_febus(fi, distance=None, time=None, attr_cls=dc.PatchAttrs):
    """Read the febus values into a patch."""
    out = []
    for attr, cm, febus in _yield_attrs_coords(fi):
        data, new_cm = _get_data_new_cm(cm, febus, distance=distance, time=time)
        patch = dc.Patch(data=data, coords=new_cm, attrs=attr_cls(**attr))
        out.append(patch)
    return out



"""
Utilities for working with xml files.
"""

from xml.etree import ElementTree


def xml_to_dict(xml_string):
    """Convert a simple xml string to a dict."""
    root = ElementTree.fromstring(xml_string)
    return _element_to_dict(root)


def _element_to_dict(element):
    """
    Recursively convert an element tree into a dict.

    Note: This function is probably not general enough to handle complicated
    xml, use with caution.
    """
    # Base case: If the element has no children, return its text content
    if len(element) == 0:
        return element.text

    # Recursive case: Convert children to dictionary
    result = {}
    for child in element:
        child_value = _element_to_dict(child)
        if child.tag in result:
            if isinstance(result[child.tag], list):
                result[child.tag].append(child_value)
            else:
                result[child.tag] = [result[child.tag], child_value]
        else:
            result[child.tag] = child_value
    return result



"""Utilities for Binary."""

from __future__ import annotations

from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path

import numpy as np
import pandas as pd
from pydantic import ConfigDict
from pydantic.alias_generators import to_pascal

import dascore as dc
from dascore.core import get_coord, get_coord_manager
from dascore.utils.misc import iterate
from dascore.utils.models import BaseModel, DateTime64
from dascore.utils.pd import _model_list_to_df, adjust_segments, filter_df
from dascore.utils.time import to_float
from dascore.utils.xml import xml_to_dict

# -- Create a pydantic model for the metadata info to help keep thins organized.

DATE_TIME_PATTERN = r"\b\d{8}T\d{6}_d{6}Z\b"
STANDARD_DIMS = ("time", "distance")


class _XMLModel(BaseModel):
    """Base model which converts camel case to snake."""

    model_config = ConfigDict(
        alias_generator=to_pascal,
    )


class XMLLaserZones(_XMLModel):
    """Zones in the xml header."""

    start_channel: int
    end_channel: int
    stride: int
    number_of_channels: int


class XMLBinaryInfo(_XMLModel):
    """Base level of information about XML index file."""

    file_format: str
    date_time: DateTime64
    das_interrogator_serial: dict[str, str]
    gauge_length_m: float
    pulse_width_ns: float
    data_type: str
    number_of_lasers: int
    i_t_u_channels: dict[str, int]
    original_temporal_sampling_rate: float
    output_temporal_sampling_rate: float
    original_spatial_sampling_interval: float
    units: str
    zones: dict[str, XMLLaserZones]
    number_of_channels: int
    number_of_frames: int
    use_relative_strain: bool
    transposed_data: bool


@lru_cache
def _read_xml_metadata(path):
    """A function to read metadata from the xml file."""
    contents = xml_to_dict(path.read_bytes())
    return XMLBinaryInfo.model_validate(contents)


def _make_distance_coord(metadata: XMLLaserZones):
    """
    Make the base coordinates from the metadata.
    """
    zones = metadata.zones
    zone = next(iter(zones.values()))
    dx = metadata.original_spatial_sampling_interval
    distance = get_coord(
        start=(zone.start_channel - 1) * dx,
        stop=zone.end_channel * dx,
        step=dx,
        units=metadata.units,
    )
    return distance


def _make_time_coord(file_start_times, metadata: XMLLaserZones):
    """Create time coord for each file."""
    dt = dc.to_timedelta64(1.0 / metadata.output_temporal_sampling_rate)
    nt = metadata.number_of_frames
    for start in file_start_times:
        yield get_coord(start=start, stop=start + dt * nt, step=dt, units="s")


def _make_base_attrs_dict(metadata: XMLLaserZones):
    """
    Make the base attributes and coordinates from metadata.
    """
    zones = metadata.zones
    assert len(zones) == 1, "expecting single zone per metadata file."
    zone_name = next(iter(zones))

    ius = metadata.das_interrogator_serial
    assert len(ius) == 1, "expecting one interrogator."
    iu_name = next(iter(ius.values()))
    attrs = dict(
        pulse_width_ns=metadata.pulse_width_ns,
        gauge_length=metadata.gauge_length_m,
        instrument_id=iu_name,
        zone_name=zone_name,
    )
    return attrs


def _get_path_datetime_64(paths):
    """Get a series of datetime64 and path as index."""
    ser = pd.Series(x.name for x in paths)
    split = ser.str.split("_", expand=True)
    year = split[1].str[:4]
    month = split[1].str[4:6]
    day = split[1].str[6:8]
    hour = split[1].str[9:11]
    minute = split[1].str[11:13]
    second = split[1].str[13:15]
    frac_sec = split[2].str.split("Z", expand=True)[0]
    iso8601 = (
        year
        + "-"
        + month
        + "-"
        + day
        + "T"
        + hour
        + ":"
        + minute
        + ":"
        + second
        + "."
        + frac_sec
    )
    dt = dc.to_datetime64(iso8601.values)
    return pd.Series(dt, index=ser.values)


def _paths_to_df(paths, metadata, attr_cls):
    """Convert paths to dataframe of info."""
    attrs = _paths_to_attrs(paths=paths, metadata=metadata, attr_cls=attr_cls)
    return _model_list_to_df(attrs)


def _read_single_file(path, metadata, time, distance):
    """Read a single file into a patch."""
    assert not metadata.transposed_data, "Cant handle data transposition yet."
    attr = _paths_to_attrs(path, metadata, summarize=False)[0]
    time_coord = attr.coords["time"].to_coord()
    distance_coord = attr.coords["distance"].to_coord()
    cm = get_coord_manager(
        {"time": time_coord, "distance": distance_coord},
        dims=attr.dim_tuple,
    )
    memmap = np.memmap(path, dtype=metadata.data_type)
    size = np.prod(cm.shape)
    assert memmap.size == size, f"wrong data shape for {path}"
    data = memmap.reshape(cm.shape)
    patch = dc.Patch(
        data=data,
        dims=attr.dim_tuple,
        coords=cm,
        attrs=attr.update(coord=None),
    )
    return patch.select(time=time, distance=distance)


def _load_patches(paths, metadata, time, distance, attr_cls):
    """Load the data file or file into a patch."""
    # Fast case for single file.
    if isinstance(paths, Path) and paths.is_file():
        return _read_single_file(paths, metadata, time, distance)
    # Since there could be **MANY** files, we have to create a mini-index
    # here to determine which files to read. Under normal circumstances
    # this isn't required as a spool can manage it.
    df = _paths_to_df(paths=paths, metadata=metadata, attr_cls=attr_cls)
    df["path"] = paths
    filtered_df = df[filter_df(df, time=time, distance=distance)].pipe(
        adjust_segments, time=time, distance=distance
    )
    # Then we can just recurse into this function.
    out = [
        _load_patches(
            Path(ser["path"]),
            metadata=metadata,
            time=time,
            distance=distance,
            attr_cls=attr_cls,
        )
        for _, ser in filtered_df.iterrows()
    ]
    return out


def _paths_to_attrs(
    paths: Path | Sequence[Path],
    metadata,
    summarize=True,
    attr_cls=dc.PatchAttrs,
    extra_attrs=None,
    timestamp=None,
):
    """Convert a path to a Patch attribute."""
    extra_attrs = {} if not extra_attrs else extra_attrs
    paths = iterate(paths)
    # filter based on timestamp
    if timestamp is not None:
        ts = to_float(timestamp)
        paths = list(x for x in paths if Path(x).stat().st_mtime >= ts)
    # No data to index.
    if not len(paths):
        return []
    base_attrs = _make_base_attrs_dict(metadata)
    distance_coord = _make_distance_coord(metadata)
    dt_ser = _get_path_datetime_64(paths)
    dims = STANDARD_DIMS[::-1] if metadata.transposed_data else STANDARD_DIMS
    out = []
    for path, time_coord in zip(paths, _make_time_coord(dt_ser.values, metadata)):
        cm = get_coord_manager(
            {"time": time_coord, "distance": distance_coord},
            dims=dims,
        )
        if summarize:
            cm = cm.to_summary_dict()
        attr_dict = {**base_attrs, **extra_attrs}
        attrs = attr_cls(coords=cm, path=path, **attr_dict)
        out.append(attrs)
    return out



"""IO module for reading prodML data."""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.constants import opt_timeable_types
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader
from dascore.utils.models import UnitQuantity, UTF8Str

from .utils import _get_cf_attrs, _get_cf_coords, _get_cf_version_str


class ProdMLPatchAttrs(dc.PatchAttrs):
    """Patch attrs for ProdML."""

    pulse_width: float = np.nan
    pulse_width_units: UnitQuantity | None = None
    gauge_length: float = np.nan
    gauge_length_units: UnitQuantity | None = None
    schema_version: UTF8Str = ""


class DASHDF5(FiberIO):
    """IO Support for DASHDF5 which uses CF version 1.7."""

    name = "DASHDF5"
    preferred_extensions = ("hdf5", "h5")
    version = "1.0"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """
        Return True if file contains terra15 version 2 data else False.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        version_str = _get_cf_version_str(resource)
        if version_str:
            return self.name, version_str

    def scan(self, resource: H5Reader, **kwargs) -> list[dc.PatchAttrs]:
        """Get metadata from file."""
        coords = _get_cf_coords(resource)
        extras = {
            "path": resource.filename,
            "file_format": self.name,
            "file_version": str(self.version),
        }
        attrs = _get_cf_attrs(resource, coords, extras=extras)
        return [attrs]

    def read(
        self,
        resource: H5Reader,
        time: tuple[opt_timeable_types, opt_timeable_types] | None = None,
        channel: tuple[float | None, float | None] | None = None,
        **kwargs,
    ):
        """Read a CF file and return a Patch."""
        coords = _get_cf_coords(resource)
        coords_new, data = coords.select(
            array=resource["das"],
            time=time,
            channel=channel,
        )
        attrs = _get_cf_attrs(resource, coords_new)
        patch = dc.Patch(
            data=data, attrs=attrs, coords=coords_new, dims=coords_new.dims
        )
        return dc.spool(patch)



"""
Deprecated module for Fourier transforms. Use
[fourier](`dascore.transform.fourier`) instead.
"""

from __future__ import annotations

import warnings
from operator import mul

import numpy as np

from dascore.constants import PatchType
from dascore.core.coords import get_coord
from dascore.units import get_quantity
from dascore.utils.patch import _get_data_units_from_dims, patch_function
from dascore.utils.time import to_float
from dascore.utils.transformatter import FourierTransformatter


@patch_function()
def rfft(patch: PatchType, dim="time") -> PatchType:
    """
    Perform a real fourier transform along the specified dimension.

    DEPRECATED FUNCTION: Use [dft](`dascore.transform.fourier.dft`) instead.
    This function is not scaled as detailed in the dascore documentation.
    """
    msg = "The Patch transform rfft is deprecated. Use dft instead."
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    assert dim in patch.dims
    axis = patch.dims.index(dim)

    ft = FourierTransformatter()
    data = patch.data
    sr = 1 / to_float(patch.attrs[f"{dim}_step"])
    freqs = np.fft.rfftfreq(data.shape[axis], sr)
    new_data = np.fft.rfft(data, axis=axis)
    # get new dims and data units
    new_dims = ft.rename_dims(dim)
    new_data_units = _get_data_units_from_dims(patch, dim, mul)
    attrs = patch.attrs.update(data_units=new_data_units)
    dims = [x if i != axis else new_dims[0] for i, x in enumerate(patch.dims)]
    # get new coord
    units = get_quantity(patch.coords.coord_map[dim].units)
    coord = get_coord(data=freqs, units=None if units is None else 1 / units)
    new_coords = {new_dims[0]: coord}
    new_coords.update({x: patch.coords.get_array(x) for x in patch.dims if x != dim})
    return patch.__class__(data=new_data, coords=new_coords, dims=dims, attrs=attrs)



"""Module for testing units."""

from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import UnitError
from dascore.units import (
    Quantity,
    assert_dtype_compatible_with_units,
    convert_units,
    get_factor_and_unit,
    get_filter_units,
    get_quantity,
    get_quantity_str,
    get_unit,
    invert_quantity,
    quant_sequence_to_quant_array,
)


class TestUnitInit:
    """Ensure units can be initialized."""

    def test_time(self):
        """Tests for time units."""
        sec = dc.get_unit("s")
        assert str(sec.dimensionality) == "[time]"

    def test_invert(self):
        """Ensure a unit can be inverted."""
        unit = dc.get_unit("s")
        inverted = invert_quantity(unit)
        reverted = invert_quantity(inverted)
        assert unit == reverted
        assert invert_quantity(None) is None


class TestGetQuantStr:
    """Ensure units can be validated."""

    valid = ("m", "s", "Hz", "1/s", "1/m", "feet", "furlongs", "km", "fortnight")
    invalid = ("bob", "gerbil inches", "farsee")

    @pytest.mark.parametrize("in_str", valid)
    def test_validate_units_good_input(self, in_str):
        """Ensure units can be validated from various inputs."""
        assert get_quantity_str(in_str)

    @pytest.mark.parametrize("in_str", invalid)
    def test_validate_units_bad_input(self, in_str):
        """Ensure units can be validated from various inputs."""
        with pytest.raises(UnitError):
            get_quantity_str(in_str)

    def test_none(self):
        """Ensure none and empty str also works."""
        assert get_quantity_str(None) is None
        assert get_quantity_str("") is None

    def test_quantity(self):
        """Ensure a quantity works."""
        # with no magnitude the string should be simple units
        quant = get_quantity("m/s")
        out = get_quantity_str(quant)
        assert out == "m / s"
        # with magnitude, it should be included.
        quant = get_quantity("10 m /s")
        out = get_quantity_str(quant)
        assert "10.0" in out

    def test_timedelta_to_quantity(self):
        """Ensure a timedelta can be converted to a quantity."""
        dt = dc.to_timedelta64(20)
        quant = dc.get_quantity(dt)
        assert quant == (20 * dc.get_unit("s"))

    def test_datetime_to_quantity(self):
        """Ensure a datetime can be converted to a quantity."""
        td = dc.to_datetime64("1970-01-01T00:00:20")
        quant = dc.get_quantity(td)
        assert quant == (20 * dc.get_unit("s"))


class TestUnitAndFactor:
    """tests for returning units and scaling factor."""

    def test_quantx_units(self):
        """Tests for the quantx unit str."""
        mag, ustr = get_factor_and_unit("rad * 2pi/2^16")
        # sometimes it is "rad * π" other times "π * rad", so just use set.
        assert set(ustr) == set("rad * π")
        assert np.isclose(mag, (2 / (2**16)))

    def test_simplify_units(self):
        """Test for reducing units."""
        mag, ustr = get_factor_and_unit("rad * (km/m)", simplify=True)
        assert get_unit(ustr) == get_unit("radian")
        assert np.isclose(mag, 1000)

    def test_none(self):
        """Ensure none returns a None and string."""
        factor, unit = get_factor_and_unit(None)
        assert factor == 1
        assert unit is None

    def test_timedelta64(self):
        """Ensure timedeltas can be separated."""
        td = dc.to_timedelta64(20)
        (factor, unit) = get_factor_and_unit(td)
        assert factor == 20.00
        assert unit == "s"

    def test_datetime64(self):
        """Ensure datetime64 can be separated."""
        td = dc.to_datetime64(20)
        (factor, unit) = get_factor_and_unit(td)
        assert factor == 20.00
        assert unit == "s"


class TestGetQuantity:
    """Tests for getting a quantity."""

    def test_quantity_identity(self):
        """Get quantity should always return the same quantity."""
        quant1 = get_quantity("1/s")
        quant2 = get_quantity("1 Hz")
        assert quant1 == get_quantity(quant1)
        assert quant1 is get_quantity(quant1)
        assert quant2 == get_quantity(quant2)
        assert quant2 is get_quantity(quant2)

    def test_get_temp(self):
        """Get quantity should work with temperatures."""
        quant1 = get_quantity("degC")
        assert "°C" in str(quant1)

    def test_timedelta64(self):
        """Ensure time deltas can be converted to quantity"""
        quant = get_quantity(dc.to_timedelta64(20))
        assert quant == (20 * dc.get_unit("s"))

    def test_datetime64(self):
        """Ensure time deltas can be converted to quantity"""
        quant = get_quantity(dc.to_datetime64(20))
        assert quant == (20 * dc.get_unit("s"))


class TestConvenientImport:
    """Tests for conveniently importing units for dascore.units."""

    def test_import_common(self):
        """Ensure common units are importable."""
        from dascore.units import Hz, ft, km, m, miles

        assert m == get_quantity("m")
        assert ft == get_quantity("ft")
        assert miles == get_quantity("miles")
        assert km == get_quantity("km")
        assert Hz == get_quantity("Hz")

    def test_bad_import_error_msg(self):
        """An import error should be raised if the unit isn't valid."""
        with pytest.raises(ImportError):
            from dascore.utils import bob  # noqa


class TestGetFilterUnits:
    """Tests for getting units that can be used for filtering."""

    def test_no_units(self):
        """Tests for when no units are specified."""
        assert get_filter_units(1, 10, "m") == (1.0, 10.0)
        assert get_filter_units(None, 10, "s") == (None, 10.0)
        assert get_filter_units(1, None, "s") == (1.0, None)

    def test_filter_units(self):
        """Tests for when filter units are already those selected."""
        hz = get_unit("Hz")
        s = get_unit("s")
        assert get_filter_units(1.0 * hz, 10.0 * hz, s) == (1.0, 10.0)
        assert get_filter_units(None, 10.0 * hz, s) == (None, 10.0)
        assert get_filter_units(1.0 * hz, 10.0 * hz, s) == (1.0, 10.0)

    def test_same_units(self):
        """Tests for when filter units are already those selected."""
        s = get_unit("s")
        assert get_filter_units(1.0 * s, 10.0 * s, s) == (0.1, 1.0)
        assert get_filter_units(None, 10.0 * s, s) == (None, 0.1)
        assert get_filter_units(10.0 * s, None, s) == (0.1, None)

    def test_different_units_raises(self):
        """The units must be the same or it should raise."""
        s, hz = get_unit("s"), get_unit("Hz")

        with pytest.raises(UnitError):
            get_filter_units(1.0, 10.0 * s, s)

        with pytest.raises(UnitError):
            get_filter_units(1.0 * s, 10.0 * hz, s)

    def test_incompatible_units_raise(self):
        """The units must be the same or it should raise."""
        s, m = get_unit("s"), get_unit("m")
        match = "Cannot convert from"
        with pytest.raises(UnitError, match=match):
            get_filter_units(1.0 * s, 10.0 * s, m)

        with pytest.raises(UnitError, match=match):
            get_filter_units(1.0 * m, 10.0 * m, s)

    def test_specifying_units_unitless_dimension_raises(self):
        """Check an error is raised when units are used on a unitless dimension."""
        msg = "Cannot use units on dimension"
        m = dc.get_unit("m")
        with pytest.raises(UnitError, match=msg):
            get_filter_units(1 * m, 2 * m, None)


class TestDTypeCompatible:
    """Ensure dtype compatibility check works."""

    quants = ("degC", "m/s", get_quantity("kg"))
    non_dt_dtypes = (np.float64, np.int_, np.float32)

    def test_non_datetime(self):
        """Any non-datetime should be compatible."""
        for quant in self.quants:
            for dtype in self.non_dt_dtypes:
                out = assert_dtype_compatible_with_units(dtype, quant)
                assert isinstance(out, Quantity)

    def test_bad_dim_raises(self):
        """Ensure a bad dimension of quantity raises."""
        for quant in self.quants:
            with pytest.raises(UnitError):
                assert_dtype_compatible_with_units(np.datetime64, quant)
            with pytest.raises(UnitError):
                assert_dtype_compatible_with_units(np.timedelta64, quant)

    def test_non_s_raises(self):
        """Only 's' should work, no other increment of time."""
        match = "only allowable units are s"
        with pytest.raises(UnitError, match=match):
            assert_dtype_compatible_with_units(np.datetime64, "ms")

    def test_s_works(self):
        """Seconds should work fine."""
        out = assert_dtype_compatible_with_units(np.datetime64, "s")
        assert out == get_quantity("s")


class TestConvertUnits:
    """Test suite for converting units."""

    def test_simple(self):
        """Simple units to simple units."""
        out = convert_units(1, "m", "ft")
        assert np.isclose(out, 0.3048)

    def test_temperature(self):
        """Ensure temperature can be converted."""
        out = convert_units(1, "m", "ft")
        assert np.isclose(out, 0.3048)

    def test_convert_offset_units(self):
        """Test simple offset units."""
        array = np.arange(10)
        f_array = array * (9 / 5) + 32.0
        out = convert_units(array, from_units="degC", to_units="degF")
        assert np.allclose(f_array, out)

    def test_convert_offset_units_with_mag(self):
        """Ensure units can be converted/set for offset units when non-1 magnitudes."""
        # One non-1 quantity
        array = np.arange(10)
        f_array = 2 * array * (9 / 5) + 32.0
        out = convert_units(array, from_units="2*degC", to_units="degF")
        assert np.allclose(f_array, out)

    def test_convert_offset_units_multiple_mags(self):
        """Ensure if both units have non-1 offsets conversion still works."""
        # Multiple non-1 quants
        array = np.arange(10)
        f_array = (array * (18 / 5) + 32.0) / 2
        out = convert_units(array, from_units="2*degC", to_units="2*degF")
        assert np.allclose(f_array, out)
        # non equal quants
        f_array = (array * (9 * 2.5 / 5) + 32.0) / 6
        out = convert_units(array, from_units="2.5*degC", to_units="6*degF")
        assert np.allclose(f_array, out)

    def test_not_output_units_raises(self):
        """Ensure an error is raised if output units are None."""
        msg = "are not specified"
        with pytest.raises(UnitError, match=msg):
            convert_units(1, from_units="m", to_units=None)

    def test_array_quantity(self):
        """Test that an array quantity works."""
        array = np.arange(10) * get_quantity("m")
        out = convert_units(array, to_units="ft")
        np.allclose(array.magnitude, out * 3.28084)


class TestQuantSequenceToQuantArray:
    """Ensure we can convert a quantity sequence to an array."""

    def test_valid_sequence_same_units(self):
        """Test with a valid sequence of quantities with the same units."""
        meter = get_quantity("m")
        sequence = [1 * meter, 2 * meter, 3 * meter]
        result = quant_sequence_to_quant_array(sequence)
        expected = np.array([1, 2, 3]) * meter
        np.testing.assert_array_equal(result.magnitude, expected.magnitude)
        assert result.units == expected.units

    def test_valid_sequence_different_units(self):
        """Test sequence of quantities with compatible but different units."""
        m, cm, km = get_quantity("m"), get_quantity("cm"), get_quantity("km")

        sequence = [1 * m, 100 * cm, 0.001 * km]
        result = quant_sequence_to_quant_array(sequence)
        expected = np.array([1, 1, 1]) * m
        assert np.allclose(result.magnitude, expected.magnitude)
        assert result.units == expected.units

    def test_incompatible_units(self):
        """Test with a sequence of quantities with incompatible units."""
        sequence = [1 * get_quantity("m"), 1 * get_quantity("s")]
        msg = "Not all values in sequence have compatible units."
        with pytest.raises(UnitError, match=msg):
            quant_sequence_to_quant_array(sequence)

    def test_non_quantity_elements(self):
        """Test with a sequence containing non-quantity elements."""
        sequence = [1 * get_quantity("m"), 5]
        msg = "Not all values in sequence are quantities."
        with pytest.raises(UnitError, match=msg):
            quant_sequence_to_quant_array(sequence)

    def test_empty_sequence(self):
        """Test with an empty sequence."""
        sequence = []
        out = quant_sequence_to_quant_array(sequence)
        assert isinstance(out, Quantity)

    def test_numpy_array_input(self):
        """Test with a numpy array input."""
        sequence = np.array([1, 2, 3])
        out = quant_sequence_to_quant_array(sequence)
        assert isinstance(out, Quantity)



"""Utilities for prodML."""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.constants import VALID_DATA_TYPES
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_coord
from dascore.utils.misc import iterate, maybe_get_items, unbyte

# --- Getting format/version

EXPECTED_ATTRS = (
    "PulseRate",
    "PulseWidth",
    "NumberOfLoci",
    "schemaVersion",
    "uuid",
)


def _get_prodml_version_str(hdf_fi) -> str:
    """Return the version string for prodml file."""
    # define a few root attrs that act as a "fingerprint" for prodML files
    acquisition = hdf_fi.get("Acquisition", None)
    if acquisition is None:
        return ""
    attrs = acquisition.attrs
    is_prodml = set(EXPECTED_ATTRS).issubset(set(attrs))
    # in some prodml schemaVersion is str, in other float; this handles both.
    version_str = str(unbyte(attrs["schemaVersion"]))
    return version_str if is_prodml else ""


def _get_raw_node_dict(acquisition_node):
    """Get a dict of {Raw[x]: node}."""
    out = {i: v for i, v in acquisition_node.items() if i.startswith("Raw")}
    return dict(sorted(out.items()))


def _get_distance_coord(acq):
    """Get the distance ranges and spacing."""

    def get_distance_units(attrs):
        """Get distance units, works for both v2 and v3."""
        name_v2_0 = "SpatialSamplingIntervalUnit"
        name_v2_1 = "SpatialSamplingInterval.uom"
        ustr = attrs.get(name_v2_0, attrs.get(name_v2_1, ""))
        return unbyte(ustr)

    attrs = acq.attrs
    step = attrs["SpatialSamplingInterval"]
    num_dist_channels = attrs["NumberOfLoci"]
    start = attrs["StartLocusIndex"] * step
    stop = start + num_dist_channels * step
    units = get_distance_units(attrs)
    return get_coord(start=start, stop=stop, units=units, step=step)


def _get_time_coord(node):
    """Get the time information from a Raw node."""
    time_array = node["RawDataTime"]
    array_len = len(time_array)
    assert array_len > 0, "Missing time array in ProdML file."
    time_attrs = time_array.attrs
    start_str = unbyte(time_attrs["PartStartTime"]).split("+")[0]
    start = dc.to_datetime64(start_str.rstrip("Z"))
    end_str = unbyte(time_attrs["PartEndTime"]).split("+")[0]
    end = dc.to_datetime64(end_str.rstrip("Z"))
    step = (end - start) / (array_len - 1)
    time_coord = get_coord(start=start, stop=end + step, step=step, units="s")
    # Sometimes the "PartEndTime" can be wrong. Check for this and try to
    # compensate. See #414.
    last = np.asarray(time_array[-1:]).astype("datetime64[us]")[0]
    tc_max = np.asarray(time_coord.max()).astype("datetime64[us]")
    diff = float(np.abs((tc_max - last) / step))
    # Note: just in case the time array is not in microseconds as it should
    # be, we prefer to use the iso 8601 strings in the 'PartStartTime' attrs
    # because they are less likely to get messed up. Therefore, we only
    # correct time coordinate from time array if the values are "close" but off.
    if 0 < diff < 10:
        time_array = time_array[:].astype("datetime64[us]")
        time_coord = get_coord(data=time_array)
    return time_coord


def _get_data_unit_and_type(node):
    """Get the data type and units."""
    attrs = node.attrs
    attr_map = {
        "RawDescription": "data_type",
        "RawDataUnit": "data_units",
    }
    out = maybe_get_items(attrs, attr_map)
    if (data_type := out.get("data_type")) is not None:
        clean = data_type.lower().replace(" ", "_")
        out["data_type"] = clean if clean in VALID_DATA_TYPES else ""
    return out


def _get_prodml_attrs(fi, extras=None) -> list[dict]:
    """Scan a prodML file, return metadata."""
    _root_attrs = {
        "PulseWidth": "pulse_width",
        "PulseWidthUnits": "pulse_width_units",
        "PulseWidth.uom": "pulse_width_units",
        "PulseRate": "pulse_rate",
        "PulseRateUnit": "pulse_rate_units",
        "PulseRate.uom": "pulse_rate_units",
        "GaugeLength": "gauge_length",
        "GaugeLengthUnit": "gauge_length_units",
        "GaugeLength.uom": "gauge_length_units",
        "schemaVersion": "schema_version",
    }
    acq = fi["Acquisition"]
    base_info = maybe_get_items(acq.attrs, _root_attrs)
    d_coord = _get_distance_coord(acq)
    raw_nodes = _get_raw_node_dict(acq)

    # Iterate each raw data node. I have only ever seen 1 in a file but since
    # it is indexed like Raw[0] there might be more.
    out = []
    for node in raw_nodes.values():
        info = dict(base_info)
        t_coord = _get_time_coord(node)
        info.update(_get_data_unit_and_type(node))
        dims = _get_dims(node)
        info["dims"] = dims
        if extras is not None:
            info.update(extras)
        info["coords"] = {"time": t_coord, "distance": d_coord}
        out.append(info)
    return out


def _get_dims(node):
    """Get the dimension names in the form of a tuple."""
    # we use distance rather than locus, setup mapping to rename this.
    map_ = {"locus": "distance", "Locus": "distance", "Time": "time"}
    data_attrs = node["RawData"].attrs
    dims = unbyte(data_attrs.get("Dimensions", "time, distance"))
    if isinstance(dims, str):
        dims = dims.replace(",", " ")
        dims = tuple(map_.get(x, x) for x in dims.split())
    else:
        unbytes = [unbyte(x) for x in iterate(dims)]
        dims = tuple(map_.get(x, x) for x in unbytes)
    return dims


def _get_data_attr(attrs, node, time, distance):
    """Get a new attributes with adjusted time/distance and data array."""
    dims = _get_dims(node)
    cm = get_coord_manager(attrs["coords"], dims=dims)
    new_cm, data = cm.select(array=node["RawData"], time=time, distance=distance)
    return data, new_cm


def _read_prodml(fi, distance=None, time=None, attr_cls=dc.PatchAttrs):
    """Read the prodml values into a patch."""
    attr_list = _get_prodml_attrs(fi)
    nodes = list(_get_raw_node_dict(fi["Acquisition"]).values())
    out = []
    for attrs, node in zip(attr_list, nodes):
        data, coords = _get_data_attr(attrs, node, time, distance)
        if data.size:
            pattrs = attr_cls(**attrs)
            out.append(dc.Patch(data=data, attrs=pattrs, coords=coords))
    return out



"""Tests for detrending functions."""

from __future__ import annotations

import numpy as np


class TestDetrend:
    """Tests for detrending data."""

    def test_detrend(self, random_patch):
        """Ensure detrending removes mean."""
        new = random_patch.new(data=random_patch.data + 10)
        # perfrom detrend, ensure all mean values are close to zero
        det = new.detrend(dim="time", type="linear")
        means = np.mean(det.data, axis=det.dims.index("time"))
        assert np.allclose(means, 0)
